(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Polynomial imports RedPair UsableRules Show
begin


types ('f,'c)"inter" = "'f \<Rightarrow> nat \<Rightarrow> ('c list \<Rightarrow> 'c)"

fun interpretationOfTerm :: "('v \<Rightarrow> 'c) \<Rightarrow> ('f,'c) inter \<Rightarrow> ('f,'v) term \<Rightarrow> 'c" ("_\<guillemotleft>_,_\<guillemotright>" 50)
where "\<beta>\<guillemotleft>\<alpha>,Var x\<guillemotright> = \<beta> x" 
   |  "\<beta>\<guillemotleft>\<alpha>,Fun f ts\<guillemotright> = (\<alpha> f (length ts) (map (\<lambda> t. \<beta>\<guillemotleft>\<alpha>,t\<guillemotright>) ts))"


lemma interpretationSub[simp] : "\<beta>\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright> = (\<lambda> x. \<beta>\<guillemotleft>\<alpha>, getSub \<sigma> x\<guillemotright>\<guillemotleft>\<alpha>, t\<guillemotright>)" 
proof (cases \<sigma>)
  case (Sub sig)
  have "\<beta>\<guillemotleft>\<alpha>, t \<cdot> Sub sig\<guillemotright> = (\<lambda> x. \<beta>\<guillemotleft>\<alpha>, sig x\<guillemotright>\<guillemotleft>\<alpha>, t\<guillemotright>)" (is "_ = (?B\<guillemotleft>\<alpha>, t\<guillemotright>)")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ss)
    let ?ts = "map (\<lambda> s. s \<cdot> Sub sig) ss"
    have "(map (interpretationOfTerm \<beta> \<alpha>) (?ts))
      = (map (interpretationOfTerm ?B \<alpha>) ss)" using Fun by (induct ss, auto)
    with Fun show ?case by simp
  qed
  with Sub show ?thesis by simp
qed




definition nonnegative_cint :: "cint \<Rightarrow> bool"
where "nonnegative_cint c \<equiv> (\<exists> c'. c = Pos c') \<or> (c = Neg 0)"


fun add :: "nat \<Rightarrow> cint \<Rightarrow> nat" 
where "add y (Neg x)  = y - x"
  |   "add y (Pos x)  = y + x"

types 'f "polyinter" = "'f \<Rightarrow> nat \<Rightarrow> (nat list) \<times> cint"

fun pol :: "'f polyinter \<Rightarrow> ('f,nat) inter"
where "pol pi f i nats = (let (args,c) = pi f i in add (listsum (map (\<lambda> x. fst x * snd x) (zip args nats))) c)"



definition poloS :: "'f polyinter \<Rightarrow> ('f,'v)trel"
  where "poloS \<alpha> \<equiv> {(s,t). (\<forall> \<beta> . (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) > (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>))}"

definition poloNS :: "'f polyinter \<Rightarrow> ('f,'v)trel"
  where "poloNS \<alpha> \<equiv> {(s,t). (\<forall> \<beta> . (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>))}"

(* note that the following four lemmas are independent of polynomial
  interpretations, but can be used for arbitrary interpretations ! *)
lemma poloSN[simp] : 
  shows "SN (poloS \<alpha> :: ('f,'v)trel)" (is "SN ?P")
proof
  fix t :: "('f,'v) term"
  show "SN ?P t"
  proof (rule ccontr)
    assume "\<not> SN ?P t"
    hence "\<exists>f . f 0 = t \<and> (\<forall>i. (f i,f(Suc i)) \<in> ?P)" by auto
    from this obtain f where "f 0 = t" and steps: "(\<forall> i. (f i, f (Suc i)) \<in> ?P)" by auto
    let ?b = "(\<lambda> x. 0) :: ('v \<Rightarrow> nat)"
    have "\<forall> i. (?b\<guillemotleft>pol \<alpha>,f i\<guillemotright>) > (?b\<guillemotleft>pol \<alpha>, f (Suc i)\<guillemotright>)"
    proof 
      fix i
      from steps have "(f i, f (Suc i)) \<in> poloS \<alpha>" .. 
      thus "(?b\<guillemotleft>pol \<alpha>, f i\<guillemotright>) > (?b\<guillemotleft>pol \<alpha>, f (Suc i)\<guillemotright>)"  unfolding poloS_def 
	by auto
    qed
    from this obtain g where nSN: "\<forall> i. ((g i) :: nat) > g (Suc i)" by force
    have "{(x,y). (x :: nat) < y} = ({(x,y). (x :: nat) > y}^-1)" by auto
    with wf_less have "wf ({(x,y). (x :: nat) > y}^-1)" by simp 
    hence "SN {(x,y). (x :: nat) > y}" by (rule wf_imp_SN)
    with nSN show False by (auto simp: SN_ars_def SN_elt_def)
  qed
qed
  

lemma poloCompat[simp] : "poloS \<alpha> O poloNS \<alpha> \<subseteq> poloS \<alpha>"
proof
  fix st
  assume "st \<in> poloS \<alpha> O poloNS \<alpha>"
  thus "st \<in> poloS \<alpha>"
  proof (cases st, auto simp: poloNS_def poloS_def)
    fix t u s \<beta>
    assume "\<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) < ( \<beta>\<guillemotleft>pol \<alpha>, u\<guillemotright>)" and 
           "\<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>, u\<guillemotright>) \<le> (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)"
    hence "(\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) < (\<beta>\<guillemotleft>pol \<alpha>, u\<guillemotright>)" and
          "(\<beta>\<guillemotleft>pol \<alpha>, u\<guillemotright>) \<le> (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)" by auto
    thus "(\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) < (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)" by simp
  qed
qed

lemma poloSSub[simp] : "sub_closed (poloS \<alpha>)" 
proof (unfold sub_closed_def poloS_def, rule subsetI)
  fix stSig
  assume stSig: "stSig \<in> CS {(s,t). \<forall>\<beta>. (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) > (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)}" (is "_ \<in> CS ?P")
  thus "stSig \<in> ?P"
  proof (cases stSig)
    case (Pair sSig tSig)
    with stSig have "(sSig, tSig) \<in> CS ?P" by simp
    hence "(sSig, tSig) \<in> ?P" 
    proof 
      fix s t \<sigma>
      assume sSig: "sSig = s \<cdot> \<sigma>" and tSig: "tSig = t \<cdot> \<sigma>" and st: "(s,t) \<in> ?P"
      thus "(sSig, tSig) \<in> ?P" by auto
    qed
    with Pair show ?thesis by simp
  qed
qed

lemma poloNSSub[simp] : "sub_closed (poloNS \<alpha>)" 
proof (unfold sub_closed_def poloNS_def, rule subsetI)
  fix stSig
  assume stSig: "stSig \<in> CS {(s,t). \<forall>\<beta>. (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)}" (is "_ \<in> CS ?P")
  thus "stSig \<in> ?P"
  proof (cases stSig)
    case (Pair sSig tSig)
    with stSig have "(sSig, tSig) \<in> CS ?P" by simp
    hence "(sSig, tSig) \<in> ?P" 
    proof 
      fix s t \<sigma>
      assume sSig: "sSig = s \<cdot> \<sigma>" and tSig: "tSig = t \<cdot> \<sigma>" and st: "(s,t) \<in> ?P"
      thus "(sSig, tSig) \<in> ?P" by auto
    qed
    with Pair show ?thesis by simp
  qed
qed

lemma addMono : "n \<ge> m \<Longrightarrow> add n c \<ge> add m c"
by (cases c, auto)


lemma addMonoS : 
 shows "n > m \<Longrightarrow> add n (Pos c) > add m (Pos c)"
by (cases c, auto)

(* to obtain weak monotonicity we indeed need the properties of polynomials *)
lemma poloCC[simp] : 
  shows "one_step_ctxt_closed (poloNS \<alpha>  :: ('f,'v)trel)" 
proof (unfold one_step_ctxt_closed_def poloNS_def, (rule allI)+, rule impI)
  fix f :: 'f 
    and bef :: "('f,'v) term list" 
    and s :: "('f,'v) term" 
    and t :: "('f,'v) term" 
    and aft :: "('f,'v) term list" 
  let ?P = "{(s,t). \<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>) \<le> (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>)}"
  assume "(s, t) \<in> ?P"
  hence st: "\<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)" by (simp)
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?P"
  proof (cases "\<alpha> f ((length bef) + 1 + (length aft))")
    case (Pair list c)
    let ?lsbb = "\<lambda> list s bef \<beta>. (\<Sum>x\<leftarrow>zip list
                 (map (interpretationOfTerm \<beta> (pol \<alpha>)) bef @
                  (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) #
                  map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)"
    have "\<forall> \<beta> . ?lsbb list s bef \<beta> \<ge> ?lsbb list t bef \<beta>" 
    proof
      fix \<beta>
      from st have stb: "(\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" .. 
      show "?lsbb list s bef \<beta> \<ge> ?lsbb list t bef \<beta>"
      proof (induct bef arbitrary: list)
	case Nil
	thus ?case using stb 
	proof (cases list, simp)
	  case (Cons n ns)
	  from stb have stbn: "n * (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> n * (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" by simp
	  from Cons have "?lsbb list s [] \<beta> = n * (\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) + (\<Sum>x\<leftarrow>zip ns  (map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)"
	    and "?lsbb list t [] \<beta> = n * (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>) + (\<Sum>x\<leftarrow>zip ns  (map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)" by auto
	  with stbn show ?thesis by arith
	qed
      next
	case (Cons b befs)
	thus ?case by (cases list, auto)
      qed
    qed
    with Pair show ?thesis by (auto simp: addMono)
  qed
qed


primrec all_positive :: "nat list \<Rightarrow> bool"
where "all_positive [] = True"
  |   "all_positive (x # xs) = ((x > 0) \<and> (all_positive xs))"

definition mono_polyinter :: "'f polyinter \<Rightarrow> bool"
where "mono_polyinter \<alpha> = (\<forall> f n. all_positive (fst (\<alpha> f n)) \<and> ((length (fst (\<alpha> f n))) = n) \<and> nonnegative_cint (snd (\<alpha> f n)))"

lemma list_len_app:
 fixes l :: "'a list" and n
 shows "(length l) = n \<Longrightarrow> 
 \<forall> k<n. \<exists> l1 l2. ((l = (l1 @ l2)) \<and> ((length l1) = k) \<and> ((length l2) = (n - k)) )"
proof(induct n arbitrary: l, simp)
 case (Suc m) 
 assume IH:"\<And> (l' :: 'a list). (length l' = m \<Longrightarrow> \<forall>k<m. \<exists>l1 l2. l' = l1 @ l2 \<and> length l1 = k \<and> length l2 = m - k)" and "length l = Suc m"
 show "\<forall>k<Suc m. \<exists>l1 l2. l = l1 @ l2 \<and> length l1 = k \<and> length l2 = Suc m - k"
  proof
   fix k
   show "k < (Suc m) \<longrightarrow> (\<exists>l1 l2. l = l1 @ l2 \<and> length l1 = k \<and> length l2 = (Suc m) - k)"
   proof(rule impI)
    assume "k < Suc m"
    thus "\<exists>l1 l2. l = l1 @ l2 \<and> length l1 = k \<and> length l2 = (Suc m) - k"
    proof(cases k)
    case 0
     from `length l = (Suc m)` have "(l = [] @ l) \<and> (length [] = 0) \<and> (length l = (Suc m) - 0)" by auto
     with 0 show "\<exists>l1 l2. l = l1 @ l2 \<and> length l1 = k \<and> length l2 = (Suc m) - k" by auto
    next
    case (Suc j)
     hence "k = Suc j" .
     from Suc `k < Suc m` have "j < m" by simp
     from `length l = (Suc m)` have "\<exists>x xs. l = x # xs \<and> length xs = m" using length_Suc_conv by auto
     then obtain x xs where "l = x # xs" and "length xs = m" by auto
     from IH `length xs = m` have "\<forall>k<m. \<exists>l1 l2. xs = l1 @ l2 \<and> length l1 = k \<and> length l2 = m - k" by auto
     with `j < m` have "\<exists>l1 l2. xs = l1 @ l2 \<and> length l1 = j \<and> length l2 = m - j" by auto
     then obtain l1 l2 where "xs = l1 @ l2 \<and> length l1 = j \<and> length l2 = m - j" by auto
     with `l = x # xs` Suc have "l = (x # l1) @ l2" and "length (x # l1) = k" and "length l2 = (Suc m) - k" by auto
     hence "l = (x # l1) @ l2 \<and> length (x # l1) = k \<and> length l2 = (Suc m) - k" by auto
     hence "\<exists>l2. l = (x # l1) @ l2 \<and> length (x # l1) = k \<and> length l2 = (Suc m) - k" by auto
     thus  "\<exists>l1 l2. l = l1 @ l2 \<and> length l1 = k \<and> length l2 = (Suc m) - k" by (rule exI)
   qed
  qed
 qed
qed

lemma poloS_CC[simp] : 
  assumes "mono_polyinter \<alpha>"
  shows "one_step_ctxt_closed (poloS \<alpha>  :: ('f,'v)trel)" 
proof (unfold one_step_ctxt_closed_def poloS_def, (rule allI)+, rule impI)
  fix f :: 'f 
    and bef :: "('f,'v) term list" 
    and s :: "('f,'v) term" 
    and t :: "('f,'v) term" 
    and aft :: "('f,'v) term list" 
  let ?P = "{(s,t). \<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>) < (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>)}"
  assume "(s, t) \<in> ?P"
  hence st: "\<forall> \<beta>. (\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) > (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)" by simp
  let ?n="(length bef) + 1 + (length aft)"
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?P"
  proof (cases "\<alpha> f ?n")
    case (Pair coeffs c)
    hence alpha_f_n_is_pair:"\<alpha> f ?n = (coeffs, c)" .
    from assms have "length (fst (\<alpha> f ?n)) = ?n" unfolding mono_polyinter_def by simp
    with Pair have "length coeffs = ?n" unfolding mono_polyinter_def by auto
    hence allk:"\<forall> k < ?n. \<exists> l1 l2. ((coeffs = (l1 @ l2)) \<and> ((length l1) = k) \<and> ((length l2) = (?n - k)) )" 
     using list_len_app[where n="?n"] by auto
    have "length bef < ?n" by auto
    with allk have "\<exists> l1 l2. ((coeffs = (l1 @ l2)) \<and> ((length l1) = (length bef)) \<and> ((length l2) = (?n - (length bef))))"  by blast
    then obtain bcoeffs acoeffs where 
     csplit:"coeffs = bcoeffs @ acoeffs" and "length bcoeffs = length bef" and "length acoeffs = 1 + length aft" by auto
    let ?lsbb = "\<lambda> list s bef \<beta>. (\<Sum>x\<leftarrow>zip list
                 (map (interpretationOfTerm \<beta> (pol \<alpha>)) bef @
                  (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) #
                  map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)"
    have allbeta:"\<forall> \<beta> . ?lsbb coeffs s bef \<beta> > ?lsbb coeffs t bef \<beta>" 
    proof
      fix \<beta>
      from st have stb: "(\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) > (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" .. thm list_induct2 
      have "?lsbb (bcoeffs @ acoeffs) s bef \<beta> > ?lsbb (bcoeffs @ acoeffs) t bef \<beta>"
      proof (rule list_induct2[where ?xs=bef and ?ys=bcoeffs])
        from `length bcoeffs = length bef` show "length bef = length bcoeffs" by simp
       next
        from `length acoeffs = 1 + length aft` have
         "\<exists> c cs. acoeffs = (c # cs)" using length_Suc_conv[where ?xs=acoeffs] by auto
        then obtain c cs where "acoeffs = (c # cs)" by auto
        from assms have allp:"all_positive  (fst (\<alpha> f ?n))" unfolding mono_polyinter_def by auto
        from alpha_f_n_is_pair allp csplit have "all_positive (bcoeffs @ acoeffs)" by auto
        hence "all_positive acoeffs" unfolding all_positive_def by (induct bcoeffs, auto)
        with `acoeffs = (c # cs)` have "c > 0"  by auto
	with stb have stbc1: "c * (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) > c * (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" by simp
	from `acoeffs = (c # cs)` 
         have "?lsbb acoeffs s [] \<beta> = c * (\<beta>\<guillemotleft>pol \<alpha>, s\<guillemotright>) + (\<Sum>x\<leftarrow>zip cs  (map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)"
	 and "?lsbb acoeffs t [] \<beta> = c * (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>) + (\<Sum>x\<leftarrow>zip cs (map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)" by auto
        with stbc1 show  "?lsbb ([] @ acoeffs) s [] \<beta> > ?lsbb ([] @ acoeffs) t [] \<beta>" by auto
       next
        fix c cs b befs
        assume same_len:"length befs = length cs" and IH:"?lsbb (cs @ acoeffs) s befs \<beta> > ?lsbb (cs @ acoeffs) t befs \<beta>"
        thus "?lsbb ((c # cs) @ acoeffs) s (b # befs) \<beta> > ?lsbb ((c # cs) @ acoeffs) t (b # befs) \<beta>"
         proof-
           have "?lsbb ((c # cs) @ acoeffs) s (b # befs) \<beta> = c * (\<beta>\<guillemotleft>pol \<alpha>, b\<guillemotright>) + (\<Sum>x\<leftarrow>zip (cs @ acoeffs)
                 (map (interpretationOfTerm \<beta> (pol \<alpha>)) befs @
                  (\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) # map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)" 
              and "?lsbb ((c # cs) @ acoeffs) t (b # befs) \<beta> = c * (\<beta>\<guillemotleft>pol \<alpha>, b\<guillemotright>) + (\<Sum>x\<leftarrow>zip (cs @ acoeffs)
                 (map (interpretationOfTerm \<beta> (pol \<alpha>)) befs @
                  (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>) # map (interpretationOfTerm \<beta> (pol \<alpha>)) aft). fst x * snd x)" by auto
           with IH show ?thesis by arith
        qed
      qed
    with `coeffs = bcoeffs @ acoeffs` show "?lsbb coeffs s bef \<beta> > ?lsbb coeffs t bef \<beta>" by auto
  qed
  from assms have "nonnegative_cint (snd (\<alpha> f ?n))" unfolding mono_polyinter_def by auto
  with Pair have "nonnegative_cint c" by auto
  hence nonneg:"(\<exists> c'. c = Pos c') \<or> (c = Neg 0)" unfolding nonnegative_cint_def by auto
  show ?thesis
  proof (cases "c = Neg 0")
   case False
    with nonneg have "\<exists> c'. c = Pos c'" by auto
    then obtain c' where "c = Pos c'" by auto
    with Pair have " \<alpha> f ?n = (coeffs, Pos c')" by auto
    with allbeta assms show ?thesis unfolding mono_polyinter_def by (auto simp: addMonoS)
   next
   case True
    with Pair have " \<alpha> f ?n = (coeffs, Neg 0)" by auto
    with allbeta assms show ?thesis unfolding mono_polyinter_def by (auto simp: addMonoS)
  qed
 qed
qed


(* theory is done, so lets prove that every polynomial order forms a reduction pair *)

interpretation polyRedpair:  redpair "poloS \<alpha>" "poloNS \<alpha>"
by (unfold_locales, auto simp: 
   poloSN 
   poloCompat 
   poloCC
   poloSSub
   poloNSSub)


(* next we need a check to test whether two terms are in relation,
   here we have to start to build polynomials  *)

datatype 'v poly = Poly "(nat \<times> 'v) list" cint (* currently only linear polys *)

instantiation poly :: (sshow) sshow
begin
fun show_pair :: "(nat \<times> 'v :: sshow) \<Rightarrow> sstring" 
where "show_pair (0,v) = []"
   |  "show_pair (Suc 0,v) = (sshow v)"
   |  "show_pair (Suc (Suc n),v) = (sshow (Suc (Suc n)) +++ ''*'' +++ sshow v)"
fun show_number :: "cint \<Rightarrow> sstring" 
where "show_number (Neg 0) = []"
    | "show_number (Pos 0) = []"
    | "show_number (Neg (Suc n)) = ''-'' @ (sshow (Suc n))"
    | "show_number (Pos (Suc n)) = sshow (Suc n)"
fun show_entries :: "('v :: sshow)poly \<Rightarrow> sstring list"
where "show_entries (Poly vp c) = (filter (\<lambda> entry. entry \<noteq> []) (show_number c # map show_pair vp))"
fun merge_entries :: "sstring list \<Rightarrow> sstring"
where "merge_entries [] = []"
    | "merge_entries (x # xs) = ('' + '' +++ x +++ (merge_entries xs))"
fun sshow_poly :: "('v :: sshow)poly \<Rightarrow> sstring"
 where "sshow_poly p = (case  show_entries p of Nil \<Rightarrow> ''0''
                             |      (Cons d es) \<Rightarrow> (d +++ (merge_entries es)))"
instance ..
end

fun addI :: "cint \<Rightarrow> cint \<Rightarrow> cint" 
where "addI (Neg x) (Neg y)  = Neg (x + y)"
  |   "addI (Pos x) (Pos y)  = Pos (x + y)"
  |   "addI (Neg x) (Pos y)  = (if x \<le> y then Pos (y - x) else Neg (x - y))"
  |   "addI (Pos x) (Neg y)  = (if y \<le> x then Pos (x - y) else Neg (y - x))"

fun mul :: "nat \<Rightarrow> cint \<Rightarrow> cint"
where "mul n (Neg m) = Neg (n*m)"
   |  "mul n (Pos m) = Pos (n*m)"

abbreviation polyZero :: "'v poly" where "polyZero \<equiv> Poly [] (Pos 0)"


fun lookupVar :: "'v \<Rightarrow> (nat \<times> 'v) list \<Rightarrow> (nat \<times> (nat \<times> 'v) list) option"
  where "lookupVar x [] = None"
    |   "lookupVar x ((c,y) # vp) = (if x = y then Some (c, vp) else 
             (case lookupVar x vp of None \<Rightarrow> None | Some (d,vpp) \<Rightarrow> Some (d, (c,y) # vpp)))"

fun sumPolyVars :: "(nat \<times> 'v) list \<Rightarrow> (nat \<times> 'v) list \<Rightarrow> (nat \<times> 'v) list"
  where "sumPolyVars [] vp2 = vp2"
     |  "sumPolyVars ((c,x) # vp1) vp2 = (case lookupVar x vp2 of 
                  None \<Rightarrow> (c,x) # sumPolyVars vp1 vp2 
                | Some (d, vpp2) \<Rightarrow> (c + d, x) # sumPolyVars vp1 vpp2)" 

fun cleanPolyVars :: "(nat \<times> 'v) list \<Rightarrow> (nat \<times> 'v) list"
where "cleanPolyVars [] = []" 
   |  "cleanPolyVars ((c,x) # vp ) = ((if c = 0 then [] else [(c,x)]) @ cleanPolyVars vp)"

fun polySum :: "'v poly \<Rightarrow> 'v poly \<Rightarrow> 'v poly" (infix "+" 65)
  where "(Poly vp1 c) + (Poly vp2 d) = Poly (cleanPolyVars (sumPolyVars vp1 vp2)) (addI c d)"

fun polyMult :: "nat \<Rightarrow> 'v poly \<Rightarrow> 'v poly" (infix "*" 70)
  where "n * (Poly vp c) = (Poly (map (\<lambda> cx. (n * fst cx, snd cx)) vp) (mul n c))"

fun polyEvalVP :: "('v \<Rightarrow> nat) \<Rightarrow> (nat \<times> 'v) list \<Rightarrow> nat" 
  where "polyEvalVP \<beta> vp = (listsum (map (\<lambda> nv. fst nv * (\<beta> (snd nv))) vp))"

lemma cleanPolyVarsSound[simp]: "polyEvalVP \<beta> (cleanPolyVars vp) = polyEvalVP \<beta> vp"
by (induct vp, auto)

fun polyEval :: "('v \<Rightarrow> nat ) \<Rightarrow> 'v poly \<Rightarrow> nat" 
  where "polyEval \<beta> (Poly vp c) = add (polyEvalVP \<beta> vp) c"

fun addAtLeastOne :: "'v poly \<Rightarrow> 'v poly" where 
  "addAtLeastOne (Poly vp (Pos cn)) = (Poly vp (Pos (Suc cn)))"
| "addAtLeastOne (Poly vp (Neg _)) = (Poly vp (Pos (Suc 0)))"
lemma addAtLeastOneSound : "polyEval \<beta> (addAtLeastOne p) \<ge> 1 + polyEval \<beta> p" 
proof (cases p)
  case (Poly vp c)
  thus ?thesis by (cases c, auto)
qed


fun polyC :: "'v poly \<Rightarrow> cint" where "polyC (Poly _ c) = c"
fun polyVP :: "'v poly \<Rightarrow> (nat \<times> 'v) list" where "polyVP (Poly vp _) = vp"


(*
locale polyorder 
begin
*)


function Pleft :: "'f polyinter \<Rightarrow> ('f,'v) term \<Rightarrow> 'v poly" 
where "Pleft \<alpha> (Var x) = (Poly [(1,x)] (Pos 0))"
   |  "Pleft \<alpha> (Fun f ts) = (let (list, c) = \<alpha> f (length ts) in 
                                  (case Poly [] c + (foldr op + (map (\<lambda> ct. (fst ct) *  (Pleft \<alpha> (snd ct))) (zip list ts)) polyZero) of
                                      Poly [] (Neg n) \<Rightarrow> polyZero | p \<Rightarrow> p))"
by (pat_completeness, auto)

termination Pleft
proof
  fix \<alpha> f and ts :: "('f,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((\<alpha>,snd ct), (\<alpha>, Fun f ts)) \<in> measures [size \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd])" by simp
qed


lemma polyMultSound[simp] : "polyEval \<beta> (c * p) = c * polyEval \<beta> p"
proof (cases p)
  case (Poly vp d)
  have "polyEval \<beta> (c * Poly vp d) = c * polyEval \<beta> (Poly vp d)" 
  proof simp
    let ?left = "\<Sum>nv\<leftarrow>map (\<lambda> cx. (c * fst cx , snd cx)) vp. fst nv * \<beta> (snd nv)"
    let ?right = "\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv)"
    have "?left = c * ?right" 
    proof (induct vp)
       case Nil thus ?case by simp
    next
      case (Cons nv vpp)
      thus ?case 
      proof (cases nv)
	case (Pair n v)
	with Cons show ?thesis by (auto simp: nat_distrib)
      qed
    qed
    thus "add ?left (mul c d) = c * (add ?right d)" by (cases d, auto simp: ring_simps nat_distrib)
  qed
  with Poly show ?thesis by simp
qed


lemma sumPolyVarsSound[simp] : "polyEvalVP \<beta> (sumPolyVars vp vq) = polyEvalVP \<beta> vp + polyEvalVP \<beta> vq"
proof simp
  show "(\<Sum>nv\<leftarrow>sumPolyVars vp vq. fst nv * \<beta> (snd nv)) =
    (\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv)) + (\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))"
  proof (induct vp arbitrary: vq)
    case Nil thus ?case by simp
  next
    case (Cons nv vpp)
    show ?case
    proof (cases nv)
      case (Pair n x)
      have "(\<Sum>nv\<leftarrow>sumPolyVars ((n,x) # vpp) vq. fst nv * \<beta> (snd nv)) =
        (\<Sum>nv\<leftarrow>(n,x) # vpp. fst nv * \<beta> (snd nv)) +
        (\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))"
      proof (cases "lookupVar x vq")
	case None
	with Cons show ?thesis by simp
      next
	case (Some evq)
	show ?thesis 
	proof (cases evq)
	  case (Pair e vqq)
	  with Some have lookup: "lookupVar x vq = Some (e, vqq)" by simp
	  with Cons show ?thesis proof (auto simp: nat_distrib)
	    from lookup show "e * \<beta> x + (\<Sum>nv\<leftarrow>vqq. fst nv * \<beta> (snd nv)) =
	      (\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))"
	    proof (induct vq arbitrary: e vqq, simp)
	      case (Cons cy vp)
	      show ?case proof (cases cy)
		case (Pair c y)
		with Cons  have lookup2: "lookupVar x ((c,y) # vp) = Some (e, vqq)" by simp
		hence "e * \<beta> x + (\<Sum>nv\<leftarrow>vqq. fst nv * \<beta> (snd nv)) = (\<Sum>nv\<leftarrow>(c,y) # vp. fst nv * \<beta> (snd nv))"
		proof (cases "x=y", simp, cases "lookupVar x vp", simp)
		  fix dvpp
		  assume xy: "x \<noteq> y" and lookup3: "lookupVar x vp = Some dvpp"
		  show ?thesis
		  proof (cases dvpp)
		    case (Pair d vpp)
		    with lookup3 have lookup4: "lookupVar x vp = Some (d, vpp)" by simp
		    with lookup2 Cons xy show ?thesis by auto
		  qed
		qed
		with Pair show ?thesis by simp
	      qed
	    qed
	  qed
	qed
      qed			    
      with Pair show ?thesis by simp
    qed
  qed
qed

declare sumPolyVars.simps polyEvalVP.simps[simp del]

lemma polySumSound1 : "polyEval \<beta> (polySum p q) \<le> polyEval \<beta> p + polyEval \<beta> q"
proof (cases p)
  case (Poly vp c)
  show ?thesis
  proof (cases q)
    case (Poly vq d)
    have "polyEval \<beta> (polySum (Poly vp c) (Poly vq d)) \<le> polyEval \<beta> (Poly vp c) + polyEval \<beta> (Poly vq d)"
      by  (cases "c", cases "d", simp, simp, cases "d", simp, simp)
    with prems show ?thesis by simp
  qed
qed

lemma PleftSound : "polyEval \<beta> (Pleft \<alpha> t) \<le> (\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>)" 
proof (induct t, simp add: polyEvalVP.simps)
  case (Fun f ts)
  show ?case 
  proof (cases "\<alpha> f (length ts)")
    case (Pair vp c)
    with Fun have "\<And>s. s \<in> set ts \<Longrightarrow> polyEval \<beta> (Pleft \<alpha> s) \<le> (\<beta>\<guillemotleft>\<lambda>a b c. split (\<lambda>args. add (\<Sum>x\<leftarrow>zip args c. fst x * snd x)) (\<alpha> a b),s\<guillemotright>)" by auto
    hence "polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pleft \<alpha> (snd ct)) (zip vp ts)) polyZero) \<le> 
      (\<Sum>x\<leftarrow>zip vp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, y) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) y))
      ts). fst x * snd x)" (is "polyEval \<beta> ?PLeftVP \<le> ?PInterVP") 
    proof (induct vp arbitrary:ts)
      case Nil
      thus ?case by (cases ts, auto simp:polyEvalVP.simps)
    next
      case (Cons n vpp)
      thus ?case proof (cases ts, simp add:polyEvalVP.simps)
	case (Cons t tss)
	from prems have ind1: "polyEval \<beta> (Pleft \<alpha> t) \<le> (\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) y,t\<guillemotright>)" by simp
	from prems have ind2: "polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pleft \<alpha> (snd ct)) (zip vpp tss)) polyZero)  \<le> 
          (\<Sum>x\<leftarrow>zip vpp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, a) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) a)) tss). fst x * snd x)" by simp
	have "polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pleft \<alpha> (snd ct)) (zip (n # vpp) (t # tss))) polyZero)
	  \<le> polyEval \<beta> (n * Pleft \<alpha> t + foldr op + (map (\<lambda>ct. fst ct * Pleft \<alpha> (snd ct)) (zip (vpp) (tss))) polyZero)" by simp
	also have "\<dots> \<le> polyEval \<beta> (n * Pleft \<alpha> t) + polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pleft \<alpha> (snd ct)) (zip (vpp) (tss))) polyZero)" by (rule polySumSound1)
	also have "\<dots> \<le> n * (\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) y,t\<guillemotright>) + polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pleft \<alpha> (snd ct)) (zip (vpp) (tss))) polyZero)" using ind1 by simp
	also have "\<dots> \<le> (\<Sum>x\<leftarrow>zip (n # vpp)
          (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, a) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) a))
          (t # tss)). fst x * snd x)" using ind2 by simp
	finally show ?thesis using Cons by simp
      qed
    qed

    with Pair Fun show ?thesis 
    proof (cases "?PLeftVP")
      case (Poly vpp d)
      with `polyEval \<beta> ?PLeftVP \<le> ?PInterVP`
      have "polyEval \<beta> (Poly vpp d) \<le> ?PInterVP" by simp
      hence part1: "polyEval \<beta> (Poly [] c + Poly vpp d)
	\<le> add ?PInterVP c"
	by (cases c, cases d, simp, simp, cases d, simp, simp)
      have part2: "polyEval \<beta> polyZero \<le> add ?PInterVP c" by (simp add: polyEvalVP.simps)
      have "polyEval \<beta>
        (case Poly [] c + Poly vpp d of 
        Poly [] (Neg n) \<Rightarrow> polyZero | Poly [] (Pos b) \<Rightarrow> Poly [] (Pos b)
        | Poly (a # lista) either \<Rightarrow> Poly (a # lista) either)
	\<le> add ?PInterVP c"
      proof (cases "Poly [] c + Poly vpp d")
	case (Poly vr e)
	thus ?thesis using part1 part2 by (cases vr, cases e, auto)
      qed
      with Poly show ?thesis using Pair Fun by simp
    qed
  qed  
qed

function Pright :: "'f polyinter \<Rightarrow> ('f,'v) term \<Rightarrow> 'v poly" 
where "Pright \<alpha> (Var x) = (Poly [(1,x)] (Pos 0))"
   |  "Pright \<alpha> (Fun f ts) = (let (list, c) = \<alpha> f (length ts) in 
                                  (case (foldr op + (map (\<lambda> ct. (fst ct) *  (Pright \<alpha> (snd ct))) (zip list ts)) (polyZero)) + (Poly [] c) of
                                      Poly vp (Pos n) \<Rightarrow> Poly vp (Pos n) | Poly vp (Neg _) \<Rightarrow> Poly vp (Pos 0)))"
by (pat_completeness, auto)

termination Pright
proof
  fix \<alpha> f and ts :: "('f,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((\<alpha>,snd ct), (\<alpha>, Fun f ts)) \<in> measures [size \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd])" by simp
qed

fun isPositive :: "cint \<Rightarrow> bool"
where "isPositive(Pos _) = True"
  |   "isPositive(Neg _) = False"

lemma mulPositive : assumes pos: "isPositive (polyC p)" shows "isPositive (polyC (n * p))"
proof (cases p)
  case (Poly vp c)
  with pos show ?thesis
    by (cases c, auto)
qed

lemma sumPositive : assumes "isPositive (polyC p)" and "isPositive(polyC q)"
  shows "isPositive(polyC (p+q))"
proof (cases p)
  case (Poly vp c)
  show ?thesis
  proof (cases q)
    case (Poly vq d)
    from prems show ?thesis by (cases "c", cases "d", auto, cases "d", auto)
  qed
qed

lemma PrightPositive : "isPositive(polyC (Pright \<alpha> t))"
proof (cases t, simp)
  case (Fun f ts)
  with assms
  show ?thesis 
  proof (cases "\<alpha> f (length ts)")
    case (Pair list c)
    with assms show ?thesis 
    proof (cases "(foldr op + (map (\<lambda> ct. (fst ct) *  (Pright \<alpha> (snd ct))) (zip list ts)) polyZero) + Poly [] c") 
      case (Poly vp d)
      with prems show ?thesis by (cases d, auto)
    qed
  qed
qed


lemma polySumSound2 : assumes 
  "isPositive(polyC p)" and "isPositive(polyC q)"
  shows "polyEval \<beta> (polySum p q) = polyEval \<beta> p + polyEval \<beta> q"
proof (cases p)
  case (Poly vp c)
  show ?thesis
  proof (cases q)
    case (Poly vq d)
    from prems show ?thesis 
      by (cases "c", auto, cases "d", auto)
  qed
qed

lemma PrightSound : "(\<beta>\<guillemotleft>pol \<alpha>, t\<guillemotright>) \<le> polyEval \<beta> (Pright \<alpha> t)"
proof (induct t, simp add: polyEvalVP.simps)
  case (Fun f ts)
  show ?case 
  proof (cases "\<alpha> f (length ts)")
    case (Pair vp c)
    with Fun have "\<And>s. s \<in> set ts \<Longrightarrow> (\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) y,s\<guillemotright>) \<le> polyEval \<beta> (Pright \<alpha> s)" by simp
    hence main: "(\<Sum>x\<leftarrow>zip vp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, y) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) y))
                        ts). fst x * snd x)
      \<le> polyEval \<beta>
      (foldr op + (map (\<lambda>ct. fst ct * Pright \<alpha> (snd ct)) (zip vp ts)) polyZero) \<and> 
      isPositive(polyC(foldr op + (map (\<lambda>ct. fst ct * Pright \<alpha> (snd ct)) (zip vp ts)) polyZero))" (is "?PInterVP \<le> polyEval \<beta> ?PRightVP \<and> _")
    proof (induct vp arbitrary:ts)
      case Nil
      thus ?case by (cases ts, auto simp:polyEvalVP.simps)
    next
      case (Cons n vpp)
      thus ?case proof (cases ts, simp add:polyEvalVP.simps)
	case (Cons t tss)
	from prems have ind1: "(\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) y,t\<guillemotright>) \<le> polyEval \<beta> (Pright \<alpha> t)" by simp
	from prems have ind2: "(\<Sum>x\<leftarrow>zip vpp (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, a) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) a)) tss). fst x * snd x) \<le> polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pright \<alpha> (snd ct)) (zip vpp tss)) polyZero) " by simp
	from prems have ind3: "isPositive(polyC (foldr op + (map (\<lambda>ct. fst ct * Pright \<alpha> (snd ct)) (zip vpp tss)) polyZero))" by simp
	from PrightPositive mulPositive have fact1: "isPositive(polyC (n * Pright \<alpha> t))" by auto
	have "(\<Sum>x\<leftarrow>zip (n # vpp)
          (map (interpretationOfTerm \<beta> (\<lambda>a b c. let (args, a) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) a))
          (t # tss)). fst x * snd x) \<le> n * (\<beta>\<guillemotleft>\<lambda>a b c. let (args, y) = \<alpha> a b in add (\<Sum>x\<leftarrow>zip args c. fst x * snd x) y,t\<guillemotright>) + polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pright \<alpha> (snd ct)) (zip (vpp) (tss))) polyZero)" using ind2 by simp
	also have "\<dots> \<le> polyEval \<beta> (n * Pright \<alpha> t) + polyEval \<beta> (foldr op + (map (\<lambda>ct. fst ct * Pright \<alpha> (snd ct)) (zip (vpp) (tss))) polyZero)" using ind1 by simp	  
	also have "\<dots> = polyEval \<beta> (n * Pright \<alpha> t + foldr op + (map (\<lambda>ct. fst ct * Pright \<alpha> (snd ct)) (zip (vpp) (tss))) polyZero)"
	  using polySumSound2[where ?p = "n * Pright \<alpha> t"] fact1 ind3 by auto
	finally show ?thesis using Cons sumPositive fact1 ind3 by auto
      qed
    qed
    show ?thesis
      (* "add ?PInterVP c
       \<le> polyEval \<beta>
          (poly_case (\<lambda>vp. cint_case (\<lambda>x. Poly vp (Pos 0)) (\<lambda>n. Poly vp (Pos n)))
            (?PRightVP + Poly [] c))" *)
    proof (cases "?PRightVP")
      case (Poly vpp d)
      from main have "?PInterVP \<le> polyEval \<beta> ?PRightVP" and "isPositive(polyC ?PRightVP)" by auto
      with Poly have "?PInterVP \<le> polyEval \<beta> (Poly vpp d)" and "isPositive(d)" by auto
      hence part1: "add ?PInterVP c \<le> polyEval \<beta> (Poly vpp d + Poly [] c)"
	by (cases d, simp, cases c, auto)
      have "add ?PInterVP c
        \<le> polyEval \<beta>
        (poly_case (\<lambda>vp. cint_case (\<lambda>x. Poly vp (Pos 0)) (\<lambda>n. Poly vp (Pos n)))
        (Poly vpp d + Poly [] c))" 
      proof (cases "Poly vpp d + Poly [] c")
	case (Poly vr e)
	with part1 show ?thesis by (cases e, auto)
      qed
      with Poly Fun Pair show ?thesis by simp
    qed
  qed
qed

declare Pright.simps Pleft.simps[simp del]

definition polyGe :: "'v poly \<Rightarrow> 'v poly \<Rightarrow> bool" (infix "\<ge>" 10)
  where "(p1 \<ge> p2) = (\<forall> \<beta>. polyEval \<beta> p1 \<ge> polyEval \<beta> p2)"
definition polyGt :: "'v poly \<Rightarrow> 'v poly \<Rightarrow> bool" (infix ">" 10)
  where "(p1 > p2) = (\<forall> \<beta>. polyEval \<beta> p1 > polyEval \<beta> p2)"


lemma polyApproxSoundGe : assumes orient: "Pleft \<alpha> s \<ge> Pright \<alpha> t"
  shows "(s,t) \<in> poloNS \<alpha>"
proof (unfold poloNS_def, simp, rule allI)
  fix \<beta> 
  from orient have "polyEval \<beta> (Pleft \<alpha> s) \<ge> polyEval \<beta> (Pright \<alpha> t)" unfolding polyGe_def by auto
  with PleftSound[where t = s and \<beta> = \<beta> and \<alpha> = \<alpha>] PrightSound[where t = t and \<beta> = \<beta> and \<alpha> = \<alpha>]
  show "(\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" by arith
qed

lemma polyApproxSoundGt : assumes orient: "Pleft \<alpha> s > Pright \<alpha> t"
  shows "(s,t) \<in> poloS \<alpha>"
proof (unfold poloS_def, simp, rule allI)
  fix \<beta> 
  from orient have "polyEval \<beta> (Pleft \<alpha> s) > polyEval \<beta> (Pright \<alpha> t)" unfolding polyGt_def by auto
  with PleftSound[where t = s and \<beta> = \<beta> and \<alpha> = \<alpha>] PrightSound[where t = t and \<beta> = \<beta> and \<alpha> = \<alpha>]
  show "(\<beta>\<guillemotleft>pol \<alpha>,s\<guillemotright>) > (\<beta>\<guillemotleft>pol \<alpha>,t\<guillemotright>)" by arith
qed


(*
end locale 
*)




(* now we come to checking that two terms are in relation  *)

(*locale polyorderImpl = polyorder
begin
*)

fun checkVPLists :: "(nat \<times> 'v) list \<Rightarrow> (nat \<times> 'v) list \<Rightarrow> 'v monad"
  where "checkVPLists _ [] = OK"
     |  "checkVPLists vp ((n,x) # vq) = (if n = 0 then checkVPLists vp vq else (case lookupVar x vp of None \<Rightarrow> error x | Some (m, vpp) \<Rightarrow> checkB (m \<ge> n) x \<guillemotright> checkVPLists vpp vq))"


(* the right polynomial will always have a positive constant by construction,
   moreover, all generated polynomials will never have Neg 0. Therefore we fail in
   all cases which are not of the form .. Pos \<ge> .. Pos *)
fun checkPoly :: "('v :: sshow) poly \<Rightarrow> 'v poly \<Rightarrow> sstring list monad"
  where "checkPoly (Poly vp c) (Poly vq d) = (
   checkB (case c of Neg _ \<Rightarrow> False | Pos cn \<Rightarrow> (case d of Pos dn \<Rightarrow> cn \<ge> dn | Neg _ \<Rightarrow> False))
      [''the constant of the left polynomial is too small'']
  \<guillemotright> checkVPLists vp vq 
     \<leftarrow> (\<lambda> x. [ ''variable '' +++ sshow x +++  '' is problematic'']))"


lemma checkPolySound[simp] : "isOK(checkPoly p q) \<Longrightarrow> p \<ge> q"
proof -
  assume check: "isOK(checkPoly p q)"
  show "p \<ge> q"
  proof (cases p)
    case (Poly vp c)
    show ?thesis
    proof (cases q)
      case (Poly vq d)
      with prems show ?thesis
      proof (cases c, simp)
	case (Pos cn)
	with prems show ?thesis 
	proof (cases d, simp)
	  case (Pos dn)
	  show ?thesis	    
	  proof (cases "cn \<ge> dn")
	    case False with prems show ?thesis by simp
	  next
	    case True      
	    { assume check2: "isOK(checkVPLists vp vq)"
	      with True 
	      have "Poly vp c \<ge> Poly  vq d" 
	      proof (unfold polyGe_def, clarify)
		fix \<beta>
		from check2 have "(\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))
		  \<le> (\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv))" 
		proof (induct vq arbitrary: vp)
		  case Nil thus ?case by simp
		next
		  case (Cons nx vqq)
		  show ?case 
		  proof (cases nx)
		    case (Pair n x)
		    with Cons show ?thesis 
		    proof (cases "n = 0", simp)
		      case False
		      with Cons Pair show ?thesis proof (cases "lookupVar x vp", simp)
			case (Some mvpp)
			show ?thesis 
			proof (cases mvpp)
			  case (Pair m vpp)
			  show ?thesis 
			  proof (cases "m \<ge> n")
			    case False with Cons Pair Some `n \<noteq> 0` `nx = (n,x)` show ?thesis by simp
			  next
			    case True with Cons Pair Some `n \<noteq> 0` `nx = (n,x)` have "isOK(checkVPLists vpp vqq)" by simp
			    with Cons have ind: "(\<Sum>nv\<leftarrow>vqq. fst nv * \<beta> (snd nv)) \<le> (\<Sum>nv\<leftarrow>vpp. fst nv * \<beta> (snd nv))" by simp
			    from Some Pair have lookup: "lookupVar x vp = Some (m,vpp)" by simp
			    have "(\<Sum>nv\<leftarrow>nx # vqq. fst nv * \<beta> (snd nv)) \<le> n * \<beta> x + (\<Sum>nv\<leftarrow> vqq. fst nv * \<beta> (snd nv))" using `nx = (n,x)` by simp
			    also have "\<dots> \<le> n * (\<beta> x) + (\<Sum>nv\<leftarrow>vpp. fst nv * \<beta> (snd nv))" using ind by simp
			    also have "\<dots> \<le> m * (\<beta> x) + (\<Sum>nv\<leftarrow>vpp. fst nv * \<beta> (snd nv))" using True by simp
			    also have "\<dots> \<le> (\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv))" (is "?left \<le> ?right") 
			    proof -
			      from lookup have "?left = ?right" 
			      proof (induct vp arbitrary: m vpp)
				case Nil thus ?case by simp
			      next
				case (Cons cy vp)
				show ?case
				proof (cases cy)
				  case (Pair c y)
				  with Cons show ?thesis proof (cases "x = y", simp)
				    case False
				    with Cons Pair show ?thesis proof (cases "lookupVar x vp", simp)
				      case (Some dvpp)
				      with False Pair Cons show ?thesis by (cases dvpp, auto)
				    qed
				  qed
				qed
			      qed
			      thus ?thesis by simp
			    qed
			    finally show ?thesis .
			  qed
			qed
		      qed
		    qed
		  qed
		qed
		with prems show "polyEval \<beta> (Poly vq d) \<le> polyEval \<beta> (Poly vp c)" by (simp add: sumPolyVars.simps  polyEvalVP.simps)
	      qed	  
	    }
	    with prems show ?thesis by simp
	  qed
	qed
      qed
    qed
  qed
qed

declare checkPoly.simps[simp del]

fun genIndices :: "nat \<Rightarrow> int \<Rightarrow> ('f,sstring) term  list"
  where "genIndices 0 _ = []" 
  |  "genIndices (Suc n) i = (Var (''x_'' +++ (sshow i)) # genIndices n (i+1))"
fun genTerm :: "'f \<Rightarrow> nat \<Rightarrow> ('f,string)term" where "genTerm f n = Fun f (genIndices n 1)"

fun "showInterpretation" :: "'f polyinter \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> sstring"
  where "showInterpretation \<alpha> f n = (sshow (Pleft \<alpha> (genTerm f n)))"


definition checkPolyGe :: "('v :: sshow) poly \<Rightarrow> ('v :: sshow ) poly \<Rightarrow> sstring list monad"
  where "checkPolyGe p q \<equiv> 
          checkPoly p q
  \<leftarrow> (\<lambda> s. (sConcat [''Could not orient '', sshow p, '' >= '', sshow q]) # s) "

definition checkPolyGt :: "('v :: sshow) poly \<Rightarrow> ('v :: sshow) poly \<Rightarrow> sstring list monad"
  where "checkPolyGt p q \<equiv> 
          checkPoly p (addAtLeastOne q)
  \<leftarrow> (\<lambda> s. (sConcat [''Could not orient '', sshow p, '' > '', sshow q]) # s) "


lemma checkPolyGeSound[simp] : "isOK(checkPolyGe p q) \<Longrightarrow> p \<ge> q" using assms unfolding checkPolyGe_def by simp
lemma checkPolyGtSound[simp] : "isOK(checkPolyGt p q) \<Longrightarrow> p > q"
using assms
proof -
  assume "isOK(checkPolyGt p q)" 
  hence ge: "p \<ge> addAtLeastOne q" unfolding checkPolyGt_def by simp
  have "\<forall> \<beta>. polyEval \<beta> p > polyEval \<beta> q" 
  proof
    fix \<beta> 
    from ge have "polyEval \<beta> p \<ge> polyEval \<beta> (addAtLeastOne q)" unfolding polyGe_def ..
    with addAtLeastOneSound[where p = q and \<beta> = \<beta>]
    show "polyEval \<beta> p > polyEval \<beta> q" by arith
  qed
  thus ?thesis by (auto simp: polyGt_def)
qed




definition checkNSPolo :: "'f polyinter \<Rightarrow> ('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad"
  where "checkNSPolo \<alpha> lr \<equiv> checkPolyGe (Pleft \<alpha>  (fst lr)) (Pright \<alpha> (snd lr)) 
      \<leftarrow> (\<lambda> s. (sConcat [ ''Could not orient '' , sshow (fst lr), '' >= '', sshow (snd lr)]) # s)
"

lemma checkNSPoloSound[simp] : "isOK(checkNSPolo \<alpha> lr) \<Longrightarrow> lr \<in> poloNS \<alpha>"
  unfolding checkNSPolo_def by (cases lr, auto simp: polyApproxSoundGe)


definition checkSPolo :: "'f polyinter \<Rightarrow> ('f :: sshow ,'v :: sshow)rule \<Rightarrow> sstring list monad"
  where "checkSPolo \<alpha> lr \<equiv> checkPolyGt (Pleft \<alpha> (fst lr)) (Pright \<alpha> (snd lr))
     \<leftarrow> (\<lambda> s. (sConcat [ ''Could not orient '' , sshow (fst lr), '' > '', sshow (snd lr)]) # s)
"

lemma checkSPoloSound[simp] : "isOK(checkSPolo \<alpha> lr) \<Longrightarrow> lr \<in> poloS \<alpha>"
  unfolding checkSPolo_def by (cases lr, auto simp: polyApproxSoundGt)


fun createRepresentation :: "('f :: sshow) polyinter \<Rightarrow> ('f \<times> nat) list \<Rightarrow> sstring list"
  where "createRepresentation \<alpha> fs =  (''polynomial interpretation: '' #
             map (\<lambda> (g,n) . sConcat [''Pol('', sshow g, ''/'', sshow n, '') = '', showInterpretation \<alpha> g n]) (remdups fs))"
(*
end 
*)


fun createPolyInter :: "('f \<times> (cint \<times> nat list)) list \<Rightarrow> 'f polyinter"
  where "createPolyInter [] f n = (replicate n 1, intToCint 1)"
     |  "createPolyInter ((g,(i,coeffs)) # rest) f n = (if (f = g) \<and> (n = length coeffs) then (coeffs, i) else createPolyInter rest f n)"

(* create an arity that is not used *)
fun createArity :: "('f \<times> (cint \<times> nat list))list \<Rightarrow> nat"
  where "createArity fcns = Suc (
            foldr (\<lambda> (g,ccs) m. max m (length (snd ccs))) fcns 0
            )"

lemma createAritySound : assumes
  m: "m \<ge> createArity l"
  shows "createPolyInter l f m = (replicate m 1, intToCint 1)"
proof -
  have n: "createArity l = Suc (foldr (\<lambda> (g,ccs) m. max m (length (snd ccs))) l 0)" by auto
  have "\<And> g i coeffs. (g, i, coeffs) \<in> set l \<Longrightarrow> length coeffs \<le> foldr (\<lambda> (g,ccs) m. max m (length (snd ccs))) l 0" 
  proof -
    fix g i coeffs
    assume "(g, i, coeffs) \<in> set l"
    thus "length coeffs \<le> foldr (\<lambda> (g,ccs) m. max m (length (snd ccs))) l 0" 
    proof (induct l, simp)
      case (Cons gic rest)
      show ?case 
      proof (cases gic)
	fix h j cs
	assume "gic = (h, j, cs)"
	with Cons show "length coeffs \<le> foldr (\<lambda> (g,ccs) m. max m (length (snd ccs))) (gic # rest) 0"
	  by (cases "gic = (g,i,coeffs)", auto)
      qed
    qed
  qed
  
  with m n have mLarge: "\<And> g i coeffs. (g, i, coeffs) \<in> set l \<Longrightarrow> length coeffs \<noteq> m" by force
  thus ?thesis
  proof (induct l)
    case Nil show ?case by simp
  next
    case (Cons gic rest)
    show ?case
    proof (cases gic)
      fix g i cs
      assume gic: "gic = (g,i,cs)"
      with Cons have "(m \<noteq> length cs)" by auto
      with gic have "createPolyInter (gic # rest) f m = createPolyInter rest f m" by auto
      also have "\<dots> = (replicate m 1, intToCint 1)" using Cons by auto
      finally show ?thesis .
    qed
  qed
qed

declare createArity.simps [simp del]

fun create_af :: "('f \<times> (cint \<times> nat list)) list \<Rightarrow> 'f af"
  where "create_af [] f n i = True"
    |   "create_af ((g,(c,coeffs)) # rest) f n i = (if (f = g) \<and> (n = length coeffs) then (coeffs ! i > 0) else create_af rest f n i)"

lemma create_af_sound : fixes l :: "('f \<times> (cint \<times> nat list)) list"
  shows "af_compatible (create_af l) (poloNS (createPolyInter l))" 
proof (unfold af_compatible_def, (rule allI)+)
  fix f :: 'f and bef and s t :: "('f,'v)term" and aft
  let ?PI = "createPolyInter l"
  let ?st = "Fun f (bef @ s # aft)"
  let ?tt = "Fun f (bef @ t # aft)"
  let ?pair = "(?st, ?tt)"
  let ?n = "Suc (length bef + length aft)"
  let ?i = "length bef"
  show "create_af l f ?n ?i \<or> ?pair \<in> poloNS ?PI"
  proof (cases "create_af l f ?n ?i", simp)
    case False
    hence zero: "(fst (?PI f ?n)) ! ?i = 0 \<and> length (fst (?PI f ?n)) = ?n"
    proof (induct l, simp)
      case (Cons gcc rest)
      show ?case 
      proof (cases gcc)
	fix g c coeffs
	assume gcc: "gcc = (g,(c,coeffs))"	
	with Cons show ?thesis by (cases "f = g \<and> ?n = length coeffs", auto)
      qed
    qed
    have "\<forall> \<beta>. (\<beta>\<guillemotleft>pol ?PI,?st\<guillemotright>) \<ge> (\<beta>\<guillemotleft>pol ?PI,?tt\<guillemotright>)" 
    proof
      fix \<beta>
      let ?in = "\<lambda> t :: ('f,'v) term. (\<beta>\<guillemotleft>pol ?PI,t\<guillemotright>)"
      let ?inter = "pol ?PI f ?n"
      have "?in ?st = ?inter (map ?in bef @ ?in s # map ?in aft)" by auto
      also have "\<dots> = ?inter (map ?in bef @ ?in t # map ?in aft)" 
      proof -
	obtain PI where PI: "PI = ?PI" by auto
	obtain beff where beff: "beff = map ?in bef" by auto
	obtain aftt where aftt: "aftt = map ?in aft" by auto
	obtain n where n: "n = ?n" by auto
	{
	  fix si ti :: "nat"
	  from zero PI beff aftt n have zeroLen: "fst (PI f n) ! (length beff) = 0 \<and> length (fst (PI f n)) = Suc (length beff + length aftt)" by auto
	  have "pol PI f n (beff @ si # aftt) = pol PI f n (beff @ ti # aftt)" 
	  proof (cases "PI f n")
	    case (Pair coeffs c)
	    hence "pol PI f n (beff @ si # aftt) = add (\<Sum>x\<leftarrow>zip coeffs (beff @ si # aftt). fst x * snd x) c" by simp
	    also have "\<dots> = add (\<Sum>x\<leftarrow>zip coeffs (beff @ ti # aftt). fst x * snd x) c" 
	    proof -
	      from zeroLen Pair have zeroLen2: "coeffs ! length beff = 0 \<and> length coeffs = Suc (length beff + length aftt)" by auto
	      hence coeffs: "coeffs = take (length beff) coeffs @ 0 # drop (Suc (length beff)) coeffs" (is "_ = ?coeffs")
		using id_take_nth_drop[where xs = coeffs and i = "length beff"] by auto
	      hence "(\<Sum>x\<leftarrow>zip coeffs (beff @ ti # aftt). fst x * snd x) = 
		(\<Sum>x\<leftarrow>zip ?coeffs (beff @ ti # aftt). fst x * snd x)" by auto
	      also have "\<dots> = (\<Sum>x\<leftarrow>zip ?coeffs (beff @ si # aftt). fst x * snd x)" using zeroLen2 by auto
	      also have "\<dots> = (\<Sum>x\<leftarrow>zip coeffs (beff @ si # aftt). fst x * snd x)" using coeffs by auto
	      finally show ?thesis by auto
	    qed		
	    also have "\<dots> = pol PI f n (beff @ ti # aftt)" using Pair by simp
	    finally show ?thesis .
	  qed
	}
	with PI beff aftt n show ?thesis by auto
      qed
      also have "\<dots> = ?in ?tt" by auto
      finally show "?in ?st \<ge> ?in ?tt" by auto
    qed      
    hence "?pair \<in> poloNS ?PI" unfolding poloNS_def by auto
    thus ?thesis .. 
  qed
qed
  


lemma poloSCeCompat : "ceTrs (c, createArity l) \<subseteq> poloS (createPolyInter l)"
unfolding poloS_def
proof (rule subsetI, simp add: ceTrs.simps)
  fix lr
  let ?n = "createArity l"
  let ?m = "Suc (Suc ?n)"
  let ?a = "createPolyInter l"
  have "?m \<ge> ?n" by auto
  hence inter: "?a c ?m = (replicate ?m 1, intToCint 1)" by (rule createAritySound)  
  fix x
  assume " (\<exists>t s. x =
               (Fun c (t # s # replicate (createArity l) (Fun c [])), t)) \<or>
        (\<exists>t s. x =
               (Fun c (t # s # replicate (createArity l) (Fun c [])), s))" ( is "?tc \<or> ?sc")  
  thus "(\<lambda> (s,t).  \<forall>\<beta>. (\<beta>\<guillemotleft>pol ?a,t\<guillemotright>)  < (\<beta>\<guillemotleft>pol ?a,s\<guillemotright>)) x"
  proof
    assume ?tc
    from this obtain t s where id: "x = (Fun c (t # s # replicate (createArity l) (Fun c [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<beta>. (\<beta>\<guillemotleft>pol ?a,t\<guillemotright>) < (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" 
    proof
      fix \<beta>      
      show "(\<beta>\<guillemotleft>pol ?a,t\<guillemotright>) < (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" using inter by auto
    qed
    thus ?thesis using id by simp
  next
    assume ?sc
    from this obtain t s where id: "x = (Fun c (t # s # replicate (createArity l) (Fun c [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<beta>. (\<beta>\<guillemotleft>pol ?a,s\<guillemotright>) < (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" 
    proof
      fix \<beta>      
      show "(\<beta>\<guillemotleft>pol ?a,s\<guillemotright>) < (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" using inter by auto
    qed
    thus ?thesis using id by simp
  qed
qed

lemma poloNSCeCompat : "ceTrs (c, createArity l) \<subseteq> poloNS (createPolyInter l)"
unfolding poloNS_def
proof (rule subsetI, simp add: ceTrs.simps)
  fix lr
  let ?n = "createArity l"
  let ?m = "Suc (Suc ?n)"
  let ?a = "createPolyInter l"
  have "?m \<ge> ?n" by auto
  hence inter: "?a c ?m = (replicate ?m 1, intToCint 1)" by (rule createAritySound)  
  fix x
  assume " (\<exists>t s. x =
               (Fun c (t # s # replicate (createArity l) (Fun c [])), t)) \<or>
        (\<exists>t s. x =
               (Fun c (t # s # replicate (createArity l) (Fun c [])), s))" ( is "?tc \<or> ?sc")  
  thus "(\<lambda> (s,t).  \<forall>\<beta>. (\<beta>\<guillemotleft>pol ?a,t\<guillemotright>)  \<le> (\<beta>\<guillemotleft>pol ?a,s\<guillemotright>)) x"
  proof
    assume ?tc
    from this obtain t s where id: "x = (Fun c (t # s # replicate (createArity l) (Fun c [])), t)" (is "x = (?l, _)") by auto    
    have "\<forall> \<beta>. (\<beta>\<guillemotleft>pol ?a,t\<guillemotright>) \<le> (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" 
    proof
      fix \<beta>      
      show "(\<beta>\<guillemotleft>pol ?a,t\<guillemotright>) \<le> (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" using inter by auto
    qed
    thus ?thesis using id by simp
  next
    assume ?sc
    from this obtain t s where id: "x = (Fun c (t # s # replicate (createArity l) (Fun c [])), s)" (is "x = (?l, _)") by auto    
    have "\<forall> \<beta>. (\<beta>\<guillemotleft>pol ?a,s\<guillemotright>) \<le> (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" 
    proof
      fix \<beta>      
      show "(\<beta>\<guillemotleft>pol ?a,s\<guillemotright>) \<le> (\<beta>\<guillemotleft>pol ?a, ?l\<guillemotright>)" using inter by auto
    qed
    thus ?thesis using id by simp
  qed
qed

lemma poloCeCompat : "ceTrs (c, createArity l) \<subseteq> poloNS (createPolyInter l) \<inter> poloS (createPolyInter l)"
  using poloSCeCompat poloNSCeCompat by best

interpretation polyCeRedpair:  ce_redpair  "poloS (createPolyInter l)" "poloNS (createPolyInter l)" "createArity l" "c"
by (unfold_locales, auto simp: 
   poloSN 
   poloCompat 
   poloCC
   poloSSub
   poloNSSub
   poloNSCeCompat
   poloSCeCompat
   )

interpretation polyCeAfRedpair:  ce_af_redpair  "poloS (createPolyInter l)" "poloNS (createPolyInter l)" "createArity l" "c" "create_af l"
by (unfold_locales, auto simp: 
   poloSN 
   poloCompat 
   poloCC
   poloSSub
   poloNSSub
   poloNSCeCompat
   poloSCeCompat
   create_af_sound
   )


datatype 'f RedPair = Polo "('f \<times> (nat \<times> nat list)) list" | NegPolo "('f \<times> (cint \<times> nat list)) list"

fun redPairToPrePolyInter :: "'f RedPair \<Rightarrow> ('f \<times> (cint \<times> nat list)) list"
  where "redPairToPrePolyInter (Polo inter) = map (\<lambda> (f, (i,coeffs)). (f, (Pos i, coeffs))) inter" 
      | "redPairToPrePolyInter (NegPolo inter) = inter"

fun createRedPairProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createRedPairProc interpr P rP R = (let \<alpha> = createPolyInter (redPairToPrePolyInter interpr) in 
  checkRedPairProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>) 
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) (redPairToPrePolyInter interpr))))
           P rP R)" 

fun createRedPairUrProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
       ('f,'v)trsL \<Rightarrow> sstring list monad"
where "createRedPairUrProc interpr P rP R ur = (let \<alpha> = createPolyInter (redPairToPrePolyInter interpr) in 
  checkRedPairUrProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>) 
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) (redPairToPrePolyInter interpr))))
           P rP R ur)" 


fun createRedPairUrAfProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
       ('f,'v)trsL \<Rightarrow> sstring list monad"
where "createRedPairUrAfProc interpr P rP R ur = (let l = redPairToPrePolyInter interpr in (let \<alpha> = createPolyInter l in
  checkRedPairUrAfProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>) 
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) (redPairToPrePolyInter interpr))))
           P rP R ur (create_af l)))" 


lemma createRedPairProcSound[simp] : fixes P rP R
assumes check: "isOK(createRedPairProc redp P rP R)"
   shows "finite_dpp (set rP, set R) \<Longrightarrow> finite_dpp (set P, set R)"
using assms 
proof (simp add: Let_def) 
  assume fin: "finite_dpp (set rP, set R)"
  let ?succ = "poloS (createPolyInter (redPairToPrePolyInter redp))"
  let ?succsim = "poloNS (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkS = "checkSPolo (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkNS = "checkNSPolo (createPolyInter (redPairToPrePolyInter redp))"
  show ?thesis 
  proof (cases "computeSplit (\<lambda> dp. isOK(?checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check have cond:
      "set P - set NSp \<subseteq> ?succ \<and> 
      ?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ   \<and> 
      set R \<subseteq> ?succsim     \<and>
      set NSp \<subseteq> set rP"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> ?succ" by (auto simp: mem_iff)
    from fin have fin2: "finite_dpp (?NSp, set R)" unfolding finite_dpp_def by (auto simp:mem_iff)
    show "finite_dpp (set P, set R)" 
    proof (rule polyRedpair.redPairProcessorDpp)
      show "finite_dpp (?NSp, set R)" using fin2 .
    next
      from stri show "set P - ?NSp \<subseteq> ?succ" .
    next
      from cond id show "?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ" by simp
    next
      from cond show "set R \<subseteq> ?succsim" by simp
    qed
  qed
qed

declare createRedPairProc.simps[simp del]


lemma createRedPairUrProcSound[simp] : fixes P rP R ur
assumes check: "isOK(createRedPairUrProc redp P rP R ur)"
   shows "finite_dpp (set rP, set R) \<Longrightarrow> finite_dpp (set P, set R)"
using assms 
proof (simp add: Let_def) 
  assume fin: "finite_dpp (set rP, set R)"
  let ?succ = "poloS (createPolyInter (redPairToPrePolyInter redp))"
  let ?succsim = "poloNS (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkS = "checkSPolo (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkNS = "checkNSPolo (createPolyInter (redPairToPrePolyInter redp))"
  show ?thesis 
  proof (cases "computeSplit (\<lambda> dp. isOK(?checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check have cond:
      "set P - set NSp \<subseteq> ?succ \<and> 
      ?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ   \<and> 
      set ur \<subseteq> ?succsim     \<and>
      set NSp \<subseteq> set rP      \<and>
      isOK(checkUrPClosed R ur P)"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> ?succ" by (auto simp: mem_iff)
    from fin have fin2: "finite_dpp (?NSp, set R)" unfolding finite_dpp_def by (auto simp:mem_iff)
    from cond have "isOK(checkUrPClosed R ur P)" by auto
    hence closed: "urClosed (set ur) R \<and> urPClosed R (set ur) (set P)" using checkUrPClosedSound by auto
    show "finite_dpp (set P, set R)" 
    proof (rule polyCeRedpair.redPairUrProcessorDpp)
      show "finite_dpp (?NSp, set R)" using fin2 .
    next
      from stri show "set P - ?NSp \<subseteq> ?succ" .
    next
      from cond id show "?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ" by simp
    next
      from cond show "set ur \<subseteq> ?succsim" by simp
    next
      from closed show "urClosed (set ur) R" by simp
    next
      from closed show "urPClosed R (set ur) (set P)" by simp
    qed
  qed
qed
    
declare createRedPairUrProc.simps[simp del]


lemma createRedPairUrAfProcSound[simp] : fixes P rP R ur
assumes check: "isOK(createRedPairUrAfProc redp P rP R ur)"
   shows "finite_dpp (set rP, set R) \<Longrightarrow> finite_dpp (set P, set R)"
using assms 
proof (simp add: Let_def) 
  assume fin: "finite_dpp (set rP, set R)"
  let ?succ = "poloS (createPolyInter (redPairToPrePolyInter redp))"
  let ?succsim = "poloNS (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkS = "checkSPolo (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkNS = "checkNSPolo (createPolyInter (redPairToPrePolyInter redp))"
  let ?af = "create_af (redPairToPrePolyInter redp)"
  show ?thesis 
  proof (cases "computeSplit (\<lambda> dp. isOK(?checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check have cond:
      "set P - set NSp \<subseteq> ?succ \<and> 
      ?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ   \<and> 
      set ur \<subseteq> ?succsim     \<and>
      set NSp \<subseteq> set rP      \<and>
      isOK(checkUrPClosed_af R ur P ?af)"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> ?succ" by (auto simp: mem_iff)
    from fin have fin2: "finite_dpp (?NSp, set R)" unfolding finite_dpp_def by (auto simp:mem_iff)
    from cond have "isOK(checkUrPClosed_af R ur P ?af)" by auto
    hence closed: "urClosed_af (set ur) R ?af \<and> urPClosed_af R (set ur) (set P) ?af" using checkUrPClosedSound_af by auto
    show "finite_dpp (set P, set R)" 
    proof (rule polyCeAfRedpair.redPairUrAfProcessorDpp)
      show "finite_dpp (?NSp, set R)" using fin2 .
    next
      from stri show "set P - ?NSp \<subseteq> ?succ" .
    next
      from cond id show "?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ" by simp
    next
      from cond show "set ur \<subseteq> ?succsim" by simp
    next
      from closed show "urClosed_af (set ur) R ?af" by simp
    next
      from closed show "urPClosed_af R (set ur) (set P) ?af" by simp
    qed
  qed
qed
    
declare createRedPairUrAfProc.simps[simp del]





fun checkListPositive :: "nat list  \<Rightarrow> sstring list monad"
 where "checkListPositive [] = OK"
 | "(checkListPositive (n # ns)) = ((checkB (n > 0) [''Coefficient 0 - not monotone'']) \<guillemotright> (checkListPositive ns))"

fun checkPoloMono :: "('f \<times> (nat \<times> nat list)) list  \<Rightarrow> sstring list monad"
 where "(checkPoloMono []) = OK"
|  "(checkPoloMono ((_, (_, coeffs)) # \<alpha>)) = ((checkListPositive coeffs) \<guillemotright> (checkPoloMono \<alpha>))"

fun checkCintNonnegative :: "cint \<Rightarrow> sstring list monad"
 where "(checkCintNonnegative (Pos _)) = OK"
 |  "(checkCintNonnegative (Neg n)) = (checkB (n = 0) [''Negative constant.''])"

fun checkNegPoloMono :: "('f \<times> (cint \<times> nat list)) list  \<Rightarrow> sstring list monad"
 where "checkNegPoloMono [] = OK"
|  "checkNegPoloMono ((_, (c, coeffs)) # \<alpha>) = ((checkCintNonnegative c)
 \<guillemotright>  (checkListPositive coeffs)  \<guillemotright> (checkNegPoloMono \<alpha>))"

fun checkInterMono :: "'f RedPair \<Rightarrow> sstring list monad"
  where "checkInterMono (Polo \<alpha>) = (checkPoloMono \<alpha>)"
| "checkInterMono (NegPolo \<alpha>) = (checkNegPoloMono \<alpha>)"

lemma checkListImpAllPositive:
 assumes "isOK(checkListPositive cs)"
 shows "all_positive cs" unfolding all_positive_def
 using assms
proof (induct cs, simp)
 case (Cons c css)
 hence "c > 0" and "isOK(checkListPositive css)" using checkB_def by auto 
 with Cons show ?case by auto
qed

abbreviation defaultInter :: "'a \<Rightarrow> nat \<Rightarrow> nat list \<times> cint" 
 where "defaultInter \<equiv> createPolyInter []"

lemma defaultInterMono:
 fixes f n
 shows "all_positive (fst (defaultInter f n)) \<and> length (fst (defaultInter f n)) = n \<and> nonnegative_cint (snd (defaultInter f n))" 
proof-
let ?pair = "defaultInter f n"
 have onesPositive:"all_positive (replicate n 1)" by (induct n, auto)
 have onesLengthN: "length (replicate n 1) = n" by simp
 have "nonnegative_cint (snd (defaultInter f n))" using nonnegative_cint_def by auto 
 with onesPositive onesLengthN show ?thesis by auto
qed

lemma checkInterMonoSound:
  assumes "isOK(checkInterMono ip)"
  shows "mono_polyinter (createPolyInter (redPairToPrePolyInter ip))"
  unfolding mono_polyinter_def
  using assms
proof(cases ip)
  case (Polo l)
  hence "ip = Polo l" .
  with assms have "isOK(checkInterMono(Polo l))" by auto
  hence "isOK(checkPoloMono l)" using checkInterMono_def by auto
  let ?mono = "(\<lambda> i f n. all_positive (fst (i f n)) \<and> ((length (fst (i f n))) = n) \<and> nonnegative_cint (snd (i f n))) :: 
    ('a \<Rightarrow> nat \<Rightarrow> nat list \<times> cint) \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> bool"
  show "\<forall> f n. (?mono (createPolyInter (redPairToPrePolyInter ip)) f n)"
  proof (rule allI)+
    fix f n
    from `isOK(checkPoloMono l)` have "?mono (createPolyInter (redPairToPrePolyInter (Polo l))) f n"
    proof (induct l arbitrary: ip)
      case Nil
      have "redPairToPrePolyInter (Polo []) = []" by auto
      thus ?case using defaultInterMono by auto
    next
      case (Cons gip gips)
      from Cons have ok:"isOK (checkPoloMono (gip # gips))" 
        and IH: "isOK (checkPoloMono gips) \<Longrightarrow> ?mono (createPolyInter (redPairToPrePolyInter (Polo gips))) f n" by auto
      thus ?case
      proof (cases gip)
        let ?gips_poly = "map (\<lambda>(f, a). (\<lambda>(i, coeffs). (f, Pos i, coeffs)) a) gips"
        let ?poly = "createPolyInter(redPairToPrePolyInter (Polo (gip # gips)))"
        case (fields g c0 coeffs)
        hence "gip = (g, c0, coeffs)" by auto
        hence create_def:"?poly f n = (if (f = g) \<and> (n = length coeffs) then (coeffs, Pos c0) else createPolyInter ?gips_poly f n)" 
          using createPolyInter_def by auto
        show "?mono ?poly f n"
        proof (cases "f = g \<and> n = (length coeffs)")
          case True
          with create_def have "nonnegative_cint (snd (?poly f n))" unfolding nonnegative_cint_def by auto
          with ok `gip = (g, c0, coeffs)` True show ?thesis using checkListImpAllPositive by simp
        next
          case False
          from ok `gip = (g, c0, coeffs)` have "isOK(checkPoloMono gips)" by auto
          from False create_def have "?poly f n = (createPolyInter ?gips_poly f n)" by auto
          with `isOK(checkPoloMono gips)` IH show ?thesis by auto
        qed
      qed    
    qed
    with `ip = Polo l` show "?mono (createPolyInter (redPairToPrePolyInter ip)) f n" by auto
  qed
next
  case (NegPolo l)
  hence "ip = NegPolo l" .
  with assms have "isOK(checkInterMono(NegPolo l))" by auto
  hence "isOK(checkNegPoloMono l)" using checkInterMono_def by auto
  let ?mono = "(\<lambda> i f n. all_positive (fst (i f n)) \<and> ((length (fst (i f n))) = n) \<and> nonnegative_cint (snd (i f n))) :: 
    ('a \<Rightarrow> nat \<Rightarrow> nat list \<times> cint) \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> bool"
  show "\<forall> f n. ?mono (createPolyInter (redPairToPrePolyInter ip)) f n"
  proof (rule allI)+
    fix f n
    from `isOK(checkNegPoloMono l)` have "?mono (createPolyInter (redPairToPrePolyInter (NegPolo l))) f n"
    proof (induct l arbitrary: ip)
      case Nil
      have "redPairToPrePolyInter (NegPolo []) = []" by auto
      thus ?case using defaultInterMono by auto
    next
      case (Cons gip gips)
      from Cons have ok:"isOK (checkNegPoloMono (gip # gips))" 
        and IH: "isOK (checkNegPoloMono gips) \<Longrightarrow> ?mono (createPolyInter (redPairToPrePolyInter (NegPolo gips))) f n" by auto
      thus ?case
      proof (cases gip)
        let ?poly = "createPolyInter(redPairToPrePolyInter (NegPolo (gip # gips)))"
        case (fields g c0 coeffs)
        hence "gip = (g, c0, coeffs)" by auto
        hence create_def:"?poly f n = (if (f = g) \<and> (n = length coeffs) then (coeffs, c0) else createPolyInter gips f n)" 
          using createPolyInter_def by auto
        show "?mono ?poly f n"
        proof (cases "f = g \<and> n = (length coeffs)")
          case True
          with create_def ok `gip = (g, c0, coeffs)` have "isOK(checkCintNonnegative c0)" by auto
          hence nonneg_constant:"nonnegative_cint c0"
          proof (cases c0, simp)
            case (Neg n)
            with `isOK(checkCintNonnegative c0)` have "n=0" using checkCintNonnegative_def checkB_def by auto
            with Neg show "nonnegative_cint (Neg n)" using nonnegative_cint_def by auto
          next
            case (Pos n)
            thus "nonnegative_cint c0" using nonnegative_cint_def by auto 
          qed
          from ok `gip = (g, c0, coeffs)` True nonneg_constant show ?thesis using checkListImpAllPositive by simp
        next
          case False
          from ok `gip = (g, c0, coeffs)` have "isOK(checkNegPoloMono gips)" by auto
          from False create_def have "?poly f n = (createPolyInter gips f n)" by auto
          with `isOK(checkNegPoloMono gips)` IH show ?thesis by auto
        qed
      qed    
    qed
    with `ip = NegPolo l` show "?mono (createPolyInter (redPairToPrePolyInter ip)) f n" by auto
  qed
qed

fun createMonoRedPairProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createMonoRedPairProc interpr P rP R rR = (
  let \<alpha> = createPolyInter (redPairToPrePolyInter interpr) in
   ((checkInterMono interpr)
 \<guillemotright> (checkMonoRedPairProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>)  
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) (redPairToPrePolyInter interpr))))
           P rP R rR)))"

fun createMonoRedPairUrProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow>
        sstring list monad"
where "createMonoRedPairUrProc interpr P rP R rR ur = (
  let \<alpha> = createPolyInter (redPairToPrePolyInter interpr) in
   ((checkInterMono interpr)
 \<guillemotright> (checkMonoRedPairUrProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>)  
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) (redPairToPrePolyInter interpr))))
           P rP R rR ur)))"

lemma createMonoRedPairProcSound[simp] : fixes P rP R rR
assumes check: "isOK(createMonoRedPairProc redp P rP R rR)"
   shows "finite_dpp (set rP, set rR) \<Longrightarrow> finite_dpp (set P, set R)"
using assms 
proof (simp add: Let_def) 
  assume fin: "finite_dpp (set rP, set rR)"
  let ?succ = "poloS (createPolyInter (redPairToPrePolyInter redp))"
  let ?succsim = "poloNS (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkS = "checkSPolo (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkNS = "checkNSPolo (createPolyInter (redPairToPrePolyInter redp))"
  show ?thesis 
  proof (cases "computeSplit (\<lambda> dp. isOK(?checkS dp)) P")
   case (Pair Sp NSp)
    hence splitP:"computeSplit (\<lambda> dp. isOK(?checkS dp)) P = (Pair Sp NSp)" .
    show ?thesis
    proof (cases "computeSplit (\<lambda> r. isOK(?checkS r)) R")
     case (Pair Sr NSr)
      hence splitR:"computeSplit (\<lambda> r. isOK(?checkS r)) R = (Pair Sr NSr)" .
      let ?NSp = "set NSp \<inter> set rP"
      let ?NSr = "set NSr \<inter> set rR"
      from splitP check have condP:
      "set P - set NSp \<subseteq> ?succ \<and> 
      ?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ   \<and> 
      set NSp \<subseteq> set rP"
       by (auto simp: Let_def mem_iff)
      hence idP: "set NSp = ?NSp" and striP: "set P - ?NSp \<subseteq> ?succ" by (auto simp: mem_iff)
      from check have "isOK(checkInterMono(redp))" unfolding checkB_def by  (auto simp: Let_def)
      hence mono:"mono_polyinter (createPolyInter (redPairToPrePolyInter redp))" by (rule checkInterMonoSound)
      show "finite_dpp (set P, set R)" 
      proof (rule polyRedpair.monoRedPairProcessorDpp[where ?rP="?NSp" and ?rR="set rR"])
       from striP show "set P - ?NSp \<subseteq> ?succ" .
      next
       from check show "set R - set rR \<subseteq> ?succ" by (auto simp: Let_def mem_iff)
      next
       from condP idP show "?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ" by simp
      next
       from check splitR have condR:"set R - ?NSr \<subseteq> ?succ \<and> set R \<inter> ?NSr \<subseteq> ?succsim" by (auto simp: Let_def mem_iff)
       thus "set rR \<inter> set R \<subseteq> ?succsim \<union> ?succ" by (auto simp: Let_def mem_iff) 
      next
       from mono show "one_step_ctxt_closed ?succ" by (rule poloS_CC)
      next
       from check show "set rR \<subseteq> set R" by (auto simp: Let_def mem_iff)  
      next
       from fin show "finite_dpp (?NSp, set rR)" unfolding finite_dpp_def by (auto simp:mem_iff)
      qed
    qed
  qed
qed

lemma createMonoRedPairUrProcSound[simp] : fixes P rP R rR ur
  assumes check: "isOK(createMonoRedPairUrProc redp P rP R rR ur)"
  shows "finite_dpp (set rP, set rR) \<Longrightarrow> finite_dpp (set P, set R)"
  using assms 
proof (simp add: Let_def) 
  assume fin: "finite_dpp (set rP, set rR)"
  let ?succ = "poloS (createPolyInter (redPairToPrePolyInter redp))"
  let ?succsim = "poloNS (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkS = "checkSPolo (createPolyInter (redPairToPrePolyInter redp))"
  let ?checkNS = "checkNSPolo (createPolyInter (redPairToPrePolyInter redp))"
  show ?thesis 
  proof (cases "computeSplit (\<lambda> dp. isOK(?checkS dp)) P")
    case (Pair Sp NSp)
    hence splitP:"computeSplit (\<lambda> dp. isOK(?checkS dp)) P = (Pair Sp NSp)" .
    show ?thesis
    proof (cases "computeSplit (\<lambda> r. isOK(?checkS r)) ur")
      case (Pair Sr NSr)
      hence splitR:"computeSplit (\<lambda> r. isOK(?checkS r)) ur = (Pair Sr NSr)" .
      let ?NSp = "set NSp \<inter> set rP"
      let ?NSr = "set NSr \<inter> set rR"
      from splitP check have condP:
      "set P - set NSp \<subseteq> ?succ \<and> 
      ?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ   \<and> 
      set NSp \<subseteq> set rP"
	by (auto simp: Let_def mem_iff)
      hence idP: "set NSp = ?NSp" and striP: "set P - ?NSp \<subseteq> ?succ" by (auto simp: mem_iff)
      from check have urClosed: "isOK(checkUrPClosed R ur P)" by  (auto simp: Let_def)
      from check have "isOK(checkInterMono(redp))" unfolding checkB_def by  (auto simp: Let_def)
      hence mono:"mono_polyinter (createPolyInter (redPairToPrePolyInter redp))" by (rule checkInterMonoSound)
      show "finite_dpp (set P, set R)" 
      proof (rule polyCeRedpair.monoRedPairUrProcessorDpp[where ?rP="?NSp" and ?rR="set rR"])
	from striP show "set P - ?NSp \<subseteq> ?succ" .
      next
	from check show "set ur - set rR \<subseteq> ?succ" by (auto simp: Let_def mem_iff)
      next
	from condP idP show "?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ" by simp
      next
	from check splitR have condR:"set ur - ?NSr \<subseteq> ?succ \<and> set ur \<inter> ?NSr \<subseteq> ?succsim" by (auto simp: Let_def mem_iff)
	thus "set rR \<inter> set ur \<subseteq> ?succsim \<union> ?succ" by (auto simp: Let_def mem_iff) 
      next
	from mono show "one_step_ctxt_closed ?succ" by (rule poloS_CC)
      next
	from check show "set rR \<subseteq> set R" by (auto simp: Let_def mem_iff)  
      next
	from fin show "finite_dpp (?NSp, set rR)" unfolding finite_dpp_def by (auto simp:mem_iff)
      next
	from urClosed show "urClosed (set ur) R" by auto
      next
	from urClosed show "urPClosed R (set ur) (set P)" by auto
      qed
    qed
  qed
qed

declare createMonoRedPairUrProc.simps[simp del]

fun createRuleRemovalProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow>  ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createRuleRemovalProc interpr R rR = (
  let \<alpha> = createPolyInter (redPairToPrePolyInter interpr) in
   ((checkInterMono interpr)
 \<guillemotright> (checkRuleRemovalProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>)  
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) (redPairToPrePolyInter interpr))))
           R rR)))"

lemma createRuleRemovalProcSound[simp] : fixes R rR
assumes check: "isOK(createRuleRemovalProc redp R rR)"
   shows "SN (rstep (set rR)) \<Longrightarrow> SN (rstep (set R))"
using assms 
proof (simp add: Let_def) 
 assume fin: "SN (rstep (set rR))"
 let ?succ = "poloS (createPolyInter (redPairToPrePolyInter redp))"
 let ?succsim = "poloNS (createPolyInter (redPairToPrePolyInter redp))"
 from check have "isOK(checkInterMono(redp))" unfolding checkB_def by  (auto simp: Let_def)
 hence mono:"mono_polyinter (createPolyInter (redPairToPrePolyInter redp))" by (rule checkInterMonoSound)
 from mono have ccS:"one_step_ctxt_closed ?succ" by (rule poloS_CC)  
 from check have nonstrict:"set R  \<subseteq> ?succsim \<union> ?succ" by (auto simp: Let_def mem_iff)
 from check have strict:"set R - set rR \<subseteq> ?succ" by (auto simp: Let_def mem_iff)
 from check have "set rR \<subseteq> set R" by (auto simp: Let_def mem_iff)
 with ccS nonstrict strict `SN (rstep (set rR))` show ?thesis 
  using polyRedpair.ruleRemoval[where ?R="set R" and ?rR="set rR"] by simp
qed

  
end




