<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:ceta="http://cl-informatik.uibk.ac.at/software/ceta" 
    version="1.0">
    <xsl:output method="html"/>
    
    <xsl:template match="/ceta:proof">
        <xsl:variable name="mode">
            <xsl:choose>
                <xsl:when test="count(/ceta:proof/ceta:loop) + count(/ceta:proof/ceta:notWellFormed) = 0">Termination proof</xsl:when>
                <xsl:otherwise>Non-termination proof</xsl:otherwise>
            </xsl:choose>
        </xsl:variable>
        <html>
            <head>
                <title>
                    <xsl:value-of select="$mode"/>
                </title>
            </head>
            <body>
                <h1>
                    <xsl:value-of select="$mode"/>
                </h1>
                <p>
                    <xsl:apply-templates>
                        <xsl:with-param name="indent" select="1"/>
                    </xsl:apply-templates>                        
                </p>
            </body>
        </html>
    </xsl:template>    
        
    <xsl:template match="ceta:notWellFormed">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: The TRS is not well-formed</h3>
        <p>The TRS violates one of the two variable conditions. Thus, it is not terminating.</p>
    </xsl:template>
    
    
    <xsl:template match="ceta:loop">
        <xsl:param name="indent"/>
        <xsl:variable name="context" select="count(ceta:box) = 0"/>
        <xsl:variable name="subst" select="count(ceta:substitution/ceta:substEntry) &gt; 0"/>
        <h3><xsl:value-of select="$indent"/>: non-termination due to loop</h3>
        The following loop proves non-termination.            
        <p>
            <table>
                <xsl:attribute name="align">center</xsl:attribute>
                <xsl:for-each select="*">
                    <xsl:if test="position() &gt; 2">
                        <tr>
                            <xsl:choose>
                                <xsl:when test="position() = 3">
                                    <td><xsl:attribute name="align">right</xsl:attribute>t</td>
                                    <td><xsl:attribute name="align">center</xsl:attribute>=</td>
                                </xsl:when>
                                <xsl:otherwise>
                                    <td/>
                                    <td><xsl:attribute name="align">center</xsl:attribute>&#8594;</td>                                    
                                </xsl:otherwise>
                            </xsl:choose>
                            <td><xsl:attribute name="align">left</xsl:attribute><xsl:apply-templates select="."/></td>
                        </tr>
                    </xsl:if>
                </xsl:for-each>
                <tr>
                    <td/>
                    <td><xsl:attribute name="align">center</xsl:attribute>&#8594;</td>                                    
                    <td><xsl:attribute name="align">left</xsl:attribute>
                        <xsl:if test="$context">C[</xsl:if>t <xsl:if test="$subst">&#963;</xsl:if><xsl:if test="$context">]</xsl:if>
                    </td>
                </tr>
            </table>          
            <xsl:if test="$subst or $context">
                where
            </xsl:if>
            <xsl:if test="$subst">
                &#963; = <xsl:apply-templates select="ceta:substitution"/>
            </xsl:if>
            <xsl:if test="$subst and $context">
                and
            </xsl:if>            
            <xsl:if test="$context">
                C = <xsl:apply-templates select="*[2]"/>
            </xsl:if>
        </p>
    </xsl:template>
    
    <xsl:template match="ceta:type">
        <xsl:choose>
            <xsl:when test="count(ceta:linearPolynomial) &gt; 0">
                Linear polynomial
            </xsl:when>
            <xsl:otherwise>
                unknown
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    
    <xsl:template match="ceta:domain">
        <xsl:choose>
            <xsl:when test="count(ceta:naturals) &gt; 0">
                the naturals
            </xsl:when>
            <xsl:otherwise>
                an unknown domain
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    
    <xsl:template name="genVars">
        <xsl:param name="n"/>
        <xsl:choose>
            <xsl:when test="$n = 0"/>
            <xsl:when test="$n = 1">
                (<span style="color:red">x<sub>1</sub></span>)
            </xsl:when>
            <xsl:when test="$n = 2">
                (<span style="color:red">x<sub>1</sub></span>, <span style="color:red">x<sub>2</sub></span>)
            </xsl:when>
            <xsl:when test="$n = 3">
                (<span style="color:red">x<sub>1</sub></span>, <span style="color:red">x<sub>2</sub></span>, <span style="color:red">x<sub>3</sub></span>)
            </xsl:when>
            <xsl:otherwise>
                (<span style="color:red">x<sub>1</sub></span>, ..., <span style="color:red">x<sub><xsl:value-of select="$n"/></sub></span>)
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    
    <xsl:template match="ceta:linearPolynomial">
        <xsl:if test="ceta:constant/ceta:number/text() &lt; '0'">
	  max(
        </xsl:if>
        <xsl:for-each select="ceta:coefficient[ceta:natural/text() != '0']">
            <xsl:if test="position() != 1"> + </xsl:if>
                <xsl:if test="ceta:natural/text() != '1'">
                    <xsl:apply-templates select="."/>      
                </xsl:if>
                <span style="color:red">x<sub><xsl:value-of select="position()"/></sub></span>
        </xsl:for-each>        
        <xsl:if test="ceta:constant/ceta:number/text() &lt; '0'">
            <xsl:text> </xsl:text>
            <xsl:apply-templates select="ceta:constant"/>
	    , 0)
        </xsl:if>
        <xsl:if test="ceta:constant/ceta:number/text() &gt; '0'">
            <xsl:if test="count(ceta:coefficient[ceta:natural/text() != 0]) &gt; 0">
                +                
            </xsl:if>
            <xsl:apply-templates select="ceta:constant"/>
        </xsl:if>
        <xsl:if test="ceta:constant/ceta:number/text() = '0'">
            <xsl:if test="count(ceta:coefficient[ceta:natural/text() != 0]) = 0">
                <xsl:apply-templates select="ceta:constant"/>                
            </xsl:if>            
        </xsl:if>
    </xsl:template>
    <xsl:template match="ceta:constant">
        <xsl:apply-templates/>
    </xsl:template>
    <xsl:template match="ceta:coefficient">        
        <xsl:apply-templates/>
    </xsl:template>
    
    <xsl:template match="ceta:interpretation">
        <p> 
            <xsl:apply-templates select="*[1]"/> 
            interpretation over 
            <xsl:apply-templates select="*[2]"/>
            <table>
                <xsl:attribute name="align">center</xsl:attribute>
                <xsl:for-each select="ceta:interpret">
                    <tr>
                        <td><xsl:attribute name="align">right</xsl:attribute>
                            [<xsl:apply-templates select="ceta:name"/><xsl:call-template name="genVars">
                                <xsl:with-param name="n" select="ceta:arity"/>
                            </xsl:call-template>]
                        </td>
                        <td><xsl:attribute name="align">center</xsl:attribute> = </td>
                        <td><xsl:attribute name="align">left</xsl:attribute><xsl:apply-templates select="ceta:function"/></td>
                    </tr>           
                </xsl:for-each>
                <tr>
                    <td><xsl:attribute name="align">right</xsl:attribute>
                        [f(<span style="color:red">x<sub>1</sub></span>, ..., <span style="color:red">x<sub>n</sub></span>)]
                    </td>
                    <td><xsl:attribute name="align">center</xsl:attribute> = </td>
                    <td><xsl:attribute name="align">left</xsl:attribute>
                        <span style="color:red">x<sub>1</sub></span> +  ... +  <span style="color:red">x<sub>n</sub></span> + 1 
                    </td>
                    <td><xsl:attribute name="align">right</xsl:attribute>
                         for all other symbols f of arity n
                    </td>                            
                </tr>
            </table>          
        </p>
    </xsl:template>
    
    <xsl:template match="ceta:dpTrans">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: switching to dependency pairs</h3>
        The following set of initial dependency pairs has been identified.
        <xsl:apply-templates select="ceta:dps/*"/>
        <xsl:apply-templates select="*[2]">
            <xsl:with-param name="indent" select="concat($indent, '.1')"/>
        </xsl:apply-templates>
    </xsl:template>
    
    
    <xsl:template match="ceta:redPairProc">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: reduction pair processor</h3>
        Using the following reduction pair
        <xsl:apply-templates select="ceta:redPair"/>        
        one remains with the following pair(s).
        <xsl:apply-templates select="ceta:dps/*"/>
        <p>
            <xsl:apply-templates select="*[3]">
                <xsl:with-param name="indent" select="concat($indent, '.1')"/>
            </xsl:apply-templates>
        </p>
    </xsl:template>

    <xsl:template match="ceta:redPairUrProc">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: reduction pair processor (with usable rules)</h3>
        Using the following reduction pair
        <xsl:apply-templates select="ceta:redPair"/>
        where the following usable rules w.r.t. the implicit argument filter of the reduction pair have been identified
        <xsl:apply-templates select="ceta:usableRules/*"/>        
        one remains with the following pair(s).
        <xsl:apply-templates select="ceta:dps/*"/>
        <p>
            <xsl:apply-templates select="*[4]">
                <xsl:with-param name="indent" select="concat($indent, '.1')"/>
            </xsl:apply-templates>
        </p>
    </xsl:template>
    
    <xsl:template match="ceta:monoRedPairProc">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: monotonic reduction pair processor</h3>
        Using the following monotonic reduction pair
        <xsl:apply-templates select="ceta:redPair"/>               
        one remains with the following pair(s).
        <xsl:apply-templates select="ceta:dps/*"/>
        and the following rule(s).        
        <xsl:apply-templates select="ceta:trs/*"/>
        <p>
            <xsl:apply-templates select="*[4]">
                <xsl:with-param name="indent" select="concat($indent, '.1')"/>
            </xsl:apply-templates>
        </p>
    </xsl:template>

    <xsl:template match="ceta:monoRedPairUrProc">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: monotonic reduction pair processor (with usable rules)</h3>
        Using the following monotonic and Ce-compatible reduction pair
        <xsl:apply-templates select="ceta:redPair"/>               
        where the following usable rules have been identified
        <xsl:apply-templates select="ceta:usableRules/*"/>        
        one remains with the following pair(s).
        <xsl:apply-templates select="ceta:dps/*"/>
        and the following rule(s).        
        <xsl:apply-templates select="ceta:trs/*"/>
        <p>
            <xsl:apply-templates select="*[5]">
                <xsl:with-param name="indent" select="concat($indent, '.1')"/>
            </xsl:apply-templates>
        </p>
    </xsl:template>
    
    
    <xsl:template match="ceta:pIsEmpty">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: P is empty </h3>
        <p>all dependency pairs have been removed.</p>
    </xsl:template>
    
    <xsl:template match="ceta:ruleRemoval">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: rule removal processor</h3>
        Using the following monotonic reduction pair
        <xsl:apply-templates select="ceta:redPair"/>                
        one remains with the TRS consisting of the following rule(s).        
        <xsl:apply-templates select="ceta:trs/*"/>
        <p>
            <xsl:apply-templates select="*[3]">
                <xsl:with-param name="indent" select="concat($indent, '.1')"/>
            </xsl:apply-templates>
        </p>
    </xsl:template>
    
    
    <xsl:template match="ceta:rIsEmpty">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: R is empty </h3>
        <p>all rules have been removed.</p>
    </xsl:template>
    
    
    
    <xsl:template match="ceta:depGraphProc">
        <xsl:param name="indent"/>
        <h3><xsl:value-of select="$indent"/>: dependency graph processor</h3>
        <xsl:variable name="all" select="count(ceta:component)"/>
        <xsl:variable name="real" select="count(ceta:component/*) - $all"/>
        <p>The dependency pairs are split into <xsl:value-of select="$real"/> component(s).</p>
        <xsl:choose>
            <xsl:when test="$real &gt; 0">
                <p><ul>    
                    <xsl:apply-templates select="." mode="iterate">
                        <xsl:with-param name="count" select="1"/>
                        <xsl:with-param name="indent" select="$indent"/>
                        <xsl:with-param name="index" select="1"/>
                        <xsl:with-param name="n" select="$all"/>
                    </xsl:apply-templates>
                </ul></p>        
            </xsl:when>
        </xsl:choose>        
    </xsl:template>
    
    <xsl:template mode="iterate" match="ceta:depGraphProc">
        <xsl:param name="indent"/>
        <xsl:param name="count"/>
        <xsl:param name="index"/>
        <xsl:param name="n"/>
        <xsl:variable name="newindex" select="$index + count(ceta:component[$count]/*) -1"/>
        <xsl:if test="$index != $newindex">
            <li>
                The
                <xsl:choose>
                    <xsl:when test="$index = 1">1st</xsl:when>
                    <xsl:when test="$index = 2">2nd</xsl:when>
                    <xsl:when test="$index = 3">3rd</xsl:when>
                    <xsl:otherwise><xsl:value-of select="$index"/>th</xsl:otherwise>
                </xsl:choose>
                component contains the
                pair(s) 
                <xsl:apply-templates select="ceta:component[$count]/ceta:dps/*"/>
                <xsl:apply-templates select="ceta:component[$count]/*[2]">
                    <xsl:with-param name="indent" select="concat($indent, '.', $index)"/>   
                </xsl:apply-templates>
            </li>
        </xsl:if>
        <xsl:if test="$count &lt; $n">
            <xsl:apply-templates select="." mode="iterate">
                <xsl:with-param name="indent" select="$indent"/>
                <xsl:with-param name="count" select="$count + 1"/>
                <xsl:with-param name="index" select="$newindex"/>
                <xsl:with-param name="n" select="$n"/>
            </xsl:apply-templates>
        </xsl:if>        
    </xsl:template>
    
    <!-- variables always in red -->
    <xsl:template name="var" match="ceta:var">
        <span style="color:red"><xsl:apply-templates/></span>
    </xsl:template>
    
    <!-- symbols in blue, sharp symbols in darkgreen-->
    <xsl:template match="ceta:name">
        <xsl:choose>
            <xsl:when test="@sharp = 'true'">
                <span style="color:darkgreen"><xsl:apply-templates/>#</span>
            </xsl:when>
            <xsl:otherwise>
                <span style="color:blue"><xsl:apply-templates/></span>
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    
    
    <xsl:template match="ceta:funapp">
        <xsl:apply-templates select="ceta:name"/>
        <xsl:for-each select="ceta:arg">
            <xsl:choose>
                <xsl:when test="position() = 1">(</xsl:when>
                <xsl:otherwise>, </xsl:otherwise>
            </xsl:choose>
            <xsl:apply-templates/>
            <xsl:if test="last() = position()">)</xsl:if>
        </xsl:for-each>
    </xsl:template>
    
    <!-- TODO: does someone know the code for box -->
    <xsl:template match="ceta:box">
        <span style="color:purple"> Hole </span>
    </xsl:template>
    
    <xsl:template match="ceta:funContext">
        <xsl:apply-templates select="ceta:name"/>
        (
        <xsl:for-each select="ceta:before/*">
            <xsl:apply-templates select="."/>,
        </xsl:for-each>
        <xsl:apply-templates select="*[3]"/>
        <xsl:for-each select="ceta:after/*">
            ,<xsl:apply-templates select="."/>
        </xsl:for-each>
        )
    </xsl:template>
    
    
    <xsl:template match="ceta:rules">
        <p>
            <xsl:choose>
                <xsl:when test="count(ceta:rule) = 0">
                    <table>
                        <xsl:attribute name="align">center</xsl:attribute>
                        <tr>
                            <td><xsl:attribute name="align">center</xsl:attribute> none </td>
                        </tr>           
                    </table>          
                </xsl:when>
                <xsl:otherwise>
                    <table>
                        <xsl:attribute name="align">center</xsl:attribute>
                        <xsl:for-each select="ceta:rule">
                            <tr>
                                <td><xsl:attribute name="align">right</xsl:attribute><xsl:apply-templates select="ceta:lhs"/></td>
                                <td><xsl:attribute name="align">center</xsl:attribute> &#8594; </td>
                                <td><xsl:attribute name="align">left</xsl:attribute><xsl:apply-templates select="ceta:rhs"/></td>
                            </tr>           
                        </xsl:for-each>
                    </table>          
                </xsl:otherwise>
            </xsl:choose>
        </p>
    </xsl:template>
    
    <xsl:template match="ceta:substitution">
        <xsl:choose>
            <xsl:when test="count(ceta:substEntry) = 0">
                [ ]
            </xsl:when>
            <xsl:otherwise>    
                [<xsl:for-each select="ceta:substEntry">
                    <xsl:apply-templates select="*[1]"/>/<xsl:apply-templates select="*[2]"/>
                    <xsl:if test="last() != position()">, </xsl:if>
                </xsl:for-each>]
            </xsl:otherwise>
        </xsl:choose>
    </xsl:template>
    
    
</xsl:stylesheet>
